import torch
import numpy as np
from scipy import ndimage
import os
from .. import  geometricTransformations,interpolation
import matplotlib.pyplot as plt
import pickle
import itertools
import scipy
from .postprocessing_utils import *
from typing import Optional, Type, Union, Tuple
from torch import Tensor
def measure_shear(device: str,workingDirectory: str,voxelToMm: Optional[list]=None,segmentsOfInterest: Optional[list]=None)->dict:
    """ Calculates the maximum shear stretch.
        :param device: sets the computation device, see torch
        :type device: string
        :param workingDirectory: path to working directory, see docs
        :type workingDirectory: string
        :param segmentsOfInterest: segmentations of interest list
        :type segmentsOfInterest: list
        :param voxelSizes: cell dimensions in mm
        :type voxelSizes: torch.Tensor
        :return shearBoundariesDict: dictionary of shear between segmentations
        :rtype shearBoundariesDict: dict
    """
    # BASICS: load images
    movingImageMask = torch.unsqueeze(torch.from_numpy(np.load(os.path.join(workingDirectory, "moving_mask.npy"))),
                                      dim=0).to(device=device)
    movingImage = torch.unsqueeze(torch.from_numpy(np.load(os.path.join(workingDirectory, "moving.npy"))), dim=0).to(
        device=device)

    indices = np.indices(movingImage.cpu()[0].size())
    pts = np.empty((np.prod(movingImage.cpu().size()), len(movingImage[0].cpu().size())))
    for i, slide in enumerate(indices):
        pts[:, i] = slide.flatten()
    pts = torch.from_numpy(pts).to(device=device).float()
    pts_orig = pts.clone()

    if segmentsOfInterest is not None:
        inter = interpolation.nearest(device,
                                      scale=torch.ones(pts.size(1), device=device))
        ptsSegmentation = inter(pts, movingImageMask).flatten().long()

        mask = torch.zeros_like(ptsSegmentation, dtype=torch.bool)
        for segment in segmentsOfInterest:
            mask |= (ptsSegmentation == segment)
        pts=pts[mask]
        ptsSegmentation=ptsSegmentation[mask]
        pts=checkAffine(device,workingDirectory,pts,ptsSegmentation)
        pts=checkNonrigid(device,workingDirectory,pts,ptsSegmentation)
        p=pts_orig.clone()
        p[mask]=pts
        pts=p
    else:
        pts = checkAffine(device, workingDirectory, pts)
        pts = checkNonrigid(device, workingDirectory, pts)

    if pts.size(1)==3:
        disX = (pts[:, 0] - pts_orig[:, 0]).flatten().reshape(movingImageMask.size())[0] * voxelToMm[0]
        disY = (pts[:, 1] - pts_orig[:, 1]).flatten().reshape(movingImageMask.size())[0] * voxelToMm[1]
        disZ = (pts[:, 2] - pts_orig[:, 2]).flatten().reshape(movingImageMask.size())[0] * voxelToMm[2]

        xdisX, ydisX, zdisX = compute_gradient_central_diff3D(disX)
        xdisY, ydisY, zdisY = compute_gradient_central_diff3D(disY)
        xdisZ, ydisZ, zdisZ = compute_gradient_central_diff3D(disZ)
        id = torch.eye(3)
        id = id.reshape((1, 3, 3))
        id = id.repeat(disX.size(0), disX.size(1), disX.size(2), 1, 1)

        graddis = torch.zeros(id.size())
        graddis[:, :, :, 0, 0] = xdisX
        graddis[:, :, :, 0, 1] = ydisX
        graddis[:, :, :, 0, 2] = zdisX
        graddis[:, :, :, 1, 0] = xdisY
        graddis[:, :, :, 1, 1] = ydisY
        graddis[:, :, :, 1, 2] = zdisY
        graddis[:, :, :, 2, 0] = xdisZ
        graddis[:, :, :, 2, 1] = ydisZ
        graddis[:, :, :, 2, 2] = zdisZ
        graddisT = torch.transpose(graddis, dim0=3, dim1=4)
        FtF = id + graddis + graddisT + torch.matmul(graddisT, graddis)
        eigenval = torch.linalg.eigvals(FtF)
        term1 = torch.sqrt(eigenval)
        # reduce imaginary to real numbers
        eigenval = torch.sqrt(term1.real ** 2 + term1.imag ** 2)
        max, _ = torch.max(eigenval, dim=3)
        min, _ = torch.min(eigenval, dim=3)
        shear = (max - min) / 2

        shearBoundariesDict={}
        shearBoundariesDict["overall shear"]=shear.flatten()

        elements = torch.unique(movingImageMask)
        result = list(itertools.combinations(elements, 2))

        for r in result:
            temp1 = torch.where(movingImageMask == r[0], 1, 0)
            temp2 = torch.where(movingImageMask == r[1], 1, 0)
            temp1 = torch.from_numpy(scipy.ndimage.binary_dilation(temp1.cpu(), iterations=1)).to(device=device).int()
            temp2 = torch.from_numpy(scipy.ndimage.binary_dilation(temp2.cpu(), iterations=1)).to(device=device).int()

            temp = temp1 + temp2
            dec = torch.where(temp == 2, 1, 0).to(device=device)
            if torch.sum(dec) > 0:
                print(f"shared boundary between {r}: ", torch.sum(dec))
                shearBoundaries = shear.flatten()[dec.flatten().cpu() == 1]
                meanShearBoundaries = torch.mean(shearBoundaries)
                stdShearBoundaries = torch.std(shearBoundaries)
                print("|mean shear boundaries| ", meanShearBoundaries, "|")
                print("|std shear boundaries |", stdShearBoundaries, "|")
                shearBoundariesDict[r]=shearBoundaries

        return shearBoundariesDict
    else:
        disX = (pts[:, 0] - pts_orig[:, 0]).flatten().reshape(movingImageMask.size())[0] * voxelToMm[0]
        disY = (pts[:, 1] - pts_orig[:, 1]).flatten().reshape(movingImageMask.size())[0] * voxelToMm[1]

        xdisX, ydisX = compute_gradient_central_diff2D(disX)
        xdisY, ydisY = compute_gradient_central_diff2D(disY)
        id = torch.eye(2)
        id = id.reshape((1, 2, 2))
        id = id.repeat(disX.size(0), disX.size(1), 1, 1)

        graddis = torch.zeros(id.size())
        graddis[:, :, 0, 0] = xdisX
        graddis[:, :, 0, 1] = ydisX
        graddis[:, :, 1, 0] = xdisY
        graddis[:, :, 1, 1] = ydisY
        graddisT = torch.transpose(graddis, dim0=2, dim1=3)
        FtF = id + graddis + graddisT + torch.matmul(graddisT, graddis)
        eigenval = torch.linalg.eigvals(FtF)
        term1 = torch.sqrt(eigenval)
        # reduce imaginary to real numbers
        eigenval = torch.sqrt(term1.real ** 2 + term1.imag ** 2)
        max, _ = torch.max(eigenval, dim=2)
        min, _ = torch.min(eigenval, dim=2)
        shear = (max - min) / 2

        shearBoundariesDict={}
        shearBoundariesDict["overall shear"]=shear.flatten()

        elements = torch.unique(movingImageMask)
        tuple_length = 2
        result = list(itertools.combinations(elements, 2))

        for r in result:
            temp1 = torch.where(movingImageMask == r[0], 1, 0)
            temp2 = torch.where(movingImageMask == r[1], 1, 0)
            temp1 = torch.from_numpy(scipy.ndimage.binary_dilation(temp1.cpu(), iterations=1)).to(device=device).int()
            temp2 = torch.from_numpy(scipy.ndimage.binary_dilation(temp2.cpu(), iterations=1)).to(device=device).int()

            temp = temp1 + temp2
            dec = torch.where(temp == 2, 1, 0).to(device=device)
            if torch.sum(dec) > 0:
                print(f"shared boundary between {r}: ", torch.sum(dec))
                shearBoundaries = shear.flatten()[dec.flatten().cpu() == 1]
                meanShearBoundaries = torch.mean(shearBoundaries)
                stdShearBoundaries = torch.std(shearBoundaries)
                print("|mean shear boundaries| ", meanShearBoundaries, "|")
                print("|std shear boundaries |", stdShearBoundaries, "|")

        return shearBoundariesDict












import torch
from typing import Optional, Type, Union, Tuple
from torch import Tensor
class l2():
    """ Class for L2 regularization of velocity fields.
    :param coef: weight of l2 norm
    :type coef: float
    :param pointsMask: a mask for pts if only a subset of pts needs to be regularized
    :type pointsMask: torch.Tensor
    :param pointsMaskLabel: the mask label for pointsMask that needs to be regularized
    :type pointsMaskLabel: int
    """
    def __init__(self,coef: float=1,pointsMask: Optional[Tensor] = None,pointsMaskLabel: Optional[Tensor] = None):
        """Constructor method."""
        self.coef=coef
        self.pointsMask=pointsMask
        self.pointsMaskLabel=pointsMaskLabel
    def __call__(self,vel: Tensor=0,**kwargs)->Tensor:
        """Calculates L2 norm.
            :param vel: point velocities
            :type vel: torch.Tensor
            :return: L2 norm
            :rtype: torch.tensor"""
        if self.pointsMask is not None:
            vel = vel[self.pointsMask == self.pointsMaskLabel]
        else:
            pass
        return self.coef*torch.sum(torch.norm(vel,dim=1))






from torch.optim.optimizer import Optimizer
import torch
from typing import Optional, Type, Union, Tuple
from torch import Tensor


class gradientDescentBacktracking(Optimizer):
    """ Class for a custom gradient descent optimizer with backtracking line search from Armijo (1966).
            :param params: FROM  PYTORCH - an iterable of torch.Tensor s or dict s. Specifies what Tensors should be optimized.
            :type params: torch.Tensor
            :param lr: learning rate
            :type lr: float
            :param max_iters: maximum iterations for the backtracking line search
            :type max_iters: int
            :param c: coefficient
            :type c: float
            :param alpha: step size reduction coefficient (0;1)
            :type alpha: float
            """
    def __init__(self, params: Tensor, lr: Optional[float]=0.1,max_iters: Optional[int]=1,c: Optional[float]=0.0004,alpha: Optional[float]=0.5):
        """ Constructor method.  """
        defaults = dict(custom_parameter=lr)
        super(gradientDescentBacktracking, self).__init__(params, defaults)
        self.loss_old=None
        self.max_iters=max_iters
        self.c=c
        self.alpha=alpha

    def step(self, closure: Type=None)->Tensor:
        """FROM PYTORCH - Performs a single optimization step.
            :param closure: closure function to calculate loss and backpropagate.
            :type closure: either a function or a class with a __call__ method
            :return: accumulated loss
            :rtype: torch.Tensor
            """
        if closure is not None:
            loss = closure()
        for group in self.param_groups:
            custom_param = group['custom_parameter']
            for p in group['params']:
                if p.grad is None:
                    continue
                gradTemp=p.grad.data.clone()
                lr=custom_param
                converged=False
                for i in range(self.max_iters):
                    p.requires_grad=False
                    p.data.add_(gradTemp, alpha=-lr)
                    lossNew = closure()
                    armijoTerm=loss.data + torch.sum(gradTemp ** 2) * lr * self.c
                    if lossNew>armijoTerm: #armijo if its inequality, see Numerical Optimization Nocedal
                        p.data.add_(gradTemp, alpha=lr)
                        lr=lr*self.alpha

                    else:
                        converged=True
                        break
                if converged:
                    p.requires_grad=True
                else:

                    p.requires_grad=True
        return loss.data





import torch
from typing import Optional, Type, Union, Tuple
from torch import Tensor
from ..interpolation import *
class similarityMeasure:
    """ Template class for image similarity measures.
    :param pts: points representing the deformable object
    :type pts: torch.Tensor
    :param intensities: intensities at material configuration
    :type intensities: torch.Tensor
    :param data: target intensity data
    :type data: torch.Tensor
    :param interpolatorIntensity: interpolator for pts locations after displacement in target intensity data
    :type interpolatorIntensity: nearest, linear, or cubic interpolation class
    :param coef: coefficient applied to the measure
    :type coef: float
    :param pointsMask: a mask for pts if only a subset of pts needs to be measured
    :type pointsMask: torch.Tensor
    :param pointsMaskLabel: the mask label for pointsMask that needs to be measured
    :type pointsMaskLabel: int
    """
    def __init__(self,pts: Tensor,intensities: Tensor,data: Tensor,interpolatorIntensity: Union[nearest,linear,cubic],coef: float=1,pointsMask: Optional[Tensor] = None,pointsMaskLabel: Optional[Tensor] = None):
        """Constructor method."""
        self.intensities=intensities
        self.data=data
        self.coef=coef
        self.interpolatorIntensity=interpolatorIntensity
        self.pts=pts
        self.pointsMask=pointsMask
        self.pointsMaskLabel=pointsMaskLabel
    def measure(self,newIntensities: Tensor,intensities: Tensor)->Tensor:
        """Method populated by children."""
        pass
    def __call__(self,dis: Tensor=0,**kwargs)->Tensor:
        """ Displaces point, feeds the measure method the coordinates of displaced points and returns the similarity between
            intensities at material coordinates vs spatial coordinates.
                :param dis: points displacements
                :type dis: torch.Tensor
                :return: similiarity measure
                :rtype: torch.Tensor
        """
        if self.pointsMask is not None:
            dis=dis[self.pointsMask==self.pointsMaskLabel]
            pts=self.pts[self.pointsMask==self.pointsMaskLabel]
            newIntensities, _, _ = self.interpolatorIntensity(pts + dis, self.data)

        else:
            newIntensities,_,_=self.interpolatorIntensity(self.pts+dis,self.data)
        return self.measure(newIntensities,self.intensities)*self.coef

class ncc(similarityMeasure):
    """ Child class for normalized cross correlation."""
    def measure(self,newIntensities: Tensor,intensities: Tensor)->Tensor:
        """Computes normalized cross correlation.
            :param newIntensities: intensities at spatial points in fixed image
            :type newIntensities: torch.Tensor
            :param val: intensities at material points in moving image
            :type val: torch.Tensor
            :return: similiarity measure.
            :rtype: torch.Tensor
        """
        pair = torch.stack([newIntensities.flatten(), intensities.flatten()], dim=1)
        mx = torch.mean(pair, dim=0)
        temp=torch.mean((pair[:, 0] - mx[0]) * (pair[:, 1] - mx[1])) / (
           torch.prod(torch.std(pair, dim=0) + torch.finfo(torch.float32).eps))
        return 10-temp

class ssd(similarityMeasure):
    """ Child class for sum of squared differences.
    """
    def measure(self,newIntensities: Tensor,intensities: Tensor)->Tensor:
        """Computes sum of squared differences.
            :param newIntensities: intensities at spatial points in fixed image
            :type newIntensities: torch.Tensor
            :param val: intensities at material points in moving image
            :type val: torch.Tensor
            :return: similiarity measure.
            :rtype: torch.Tensor
        """
        pair = torch.stack([newIntensities.flatten(), intensities.flatten()], dim=1)
        mx = torch.sum((pair[:, 0] - pair[:, 1]) ** 2)
        return mx




class nmi(similarityMeasure):
    """ Child class for mutual information.
    """
    def Histogram2D(self, vals: Tensor)->Tuple[Tensor,Tensor,Tensor]:
        """ Creates 3 histograms, 1 for the joint probability of intensity values and 2 for separate probabilities.
            intensities at original coordinates vs displaced coordinates.
                :param vals: combined intensities at material and spatial coordinates
                :type vals: torch.Tensor
                :return hist: joint probability
                :rtype hist: torch.Tensor
                :return hist_a: single probability
                :rtype hist_a: torch.Tensor
                :return hist_b: single probability
                :rtype hist_b: torch.Tensor
        """
        rangeh = torch.ceil(vals.max() - vals.min()).long()
        t_idx = vals.floor().long()
        p = torch.arange(vals.size(0))
        t = vals - t_idx
        ones4 = torch.ones([2, 2], dtype=torch.int32, device=self.device)
        onesp = torch.ones(t.size(0), dtype=torch.int32, device=self.device)
        stride_x, stride_y = torch.meshgrid(
            [torch.arange(0, 2, device=self.device) - 1, torch.arange(0, 2, device=self.device) - 1])
        t_idx = t_idx.flatten()
        indices = torch.einsum('a,bc->abc', t_idx[2 * p], ones4) * (rangeh)
        indices += torch.einsum('a,bc->abc', onesp, stride_x) * rangeh
        indices += torch.einsum('a,bc->abc', t_idx[2 * p + 1], ones4)
        indices += torch.einsum('a,bc->abc', onesp, stride_y)
        y = torch.stack([1 - t.flatten(), t.flatten()], dim=1)
        res = (torch.einsum('ab,ac->abc', y[2 * p, :], y[2 * p + 1, :]))
        v, ids = indices.flatten().unique(return_counts=True)
        val = torch.split(res.flatten(), ids.tolist());
        hist = torch.zeros(v.size(), device=self.device, dtype=torch.float32)
        va = (v % rangeh)
        vb = ((v / rangeh).long())
        for index, value in enumerate(val):
            hist[index] = value.sum()
        v_a, ids = va.unique(return_counts=True)
        hist_a = torch.zeros(v_a.size(), device=self.device, dtype=torch.float32)
        vala = torch.split(hist, ids.tolist());
        for index, value in enumerate(vala):
            hist_a[index] = value.sum()
        v_b, ids = vb.unique(return_counts=True)
        hist_b = torch.zeros(v_b.size(), device=self.device, dtype=torch.float32)
        valb = torch.split(hist, ids.tolist());
        for index, value in enumerate(valb):
            hist_b[index] = value.sum()
        hist = hist + torch.finfo(torch.float32).eps
        hist_a = hist_a + torch.finfo(torch.float32).eps
        hist_b = hist_b + torch.finfo(torch.float32).eps
        return hist, hist_a, hist_b

    def measure(self,newIntensities: Tensor,intensities: Tensor)->Tensor:
        """Computes normalized mutual information.
            :param newIntensities: intensities at spatial points in fixed image
            :type newIntensities: torch.Tensor
            :param val: intensities at material points in moving image
            :type val: torch.Tensor
            :return: similiarity measure.
            :rtype: torch.Tensor
        """
        self.device=newIntensities.device
        x = torch.stack([newIntensities.flatten(), intensities.flatten()], dim=1)
        h1, h2, h3 = self.Histogram2D(x)
        h1 = h1 / h1.sum()
        h2 = h2 / h2.sum()
        h3 = h3 / h3.sum()
        return 10 - ((torch.sum(-h2 * torch.log(h2)) + torch.sum(-h3 * torch.log(h3))) / torch.sum(-h1 * torch.log(h1)))




import torch
from typing import Optional, Type, Union, Tuple
from torch import Tensor
from ..interpolation import *
from ..numericalIntegration import *

class nonrigidDeformation():
    """Class for deformable transformation, based on velocity fields.
            :param pts: points in material configuration
            :type pts: torch.Tensor
            :param integrator: numerical integration method to solve teh initial value problem for the connecting differential equations
            :type integrator:  trapezoidal or forwardEuler class
            :param interpolator: interpolation method for interpolating velocity fields
            :type interpolator: interpolation class
    """
    def __init__(self,pts: Tensor,integrator: Union[trapezoidal,forwardEuler],interpolator: Union[nearest,linear,cubic]):
        """Constructor method."""
        self.pts=pts
        self.integrator = integrator
        self.interpolator = interpolator
    def __call__(self,velocityField: Tensor,mainTerms: Optional[list] = None,regTerms: Optional[list] = None)-> Tuple[Tensor,Tensor]:
        """Displaces points with nonrigid displacement. Evaluates mainTerm and regTerms on the fly.
        :param velocityField: set of velocity fields (time frames, dimensions, # control points dim 1, # control points dim 2 [,# control points dim 3])
        :type velocityField: torch.Tensor
        :param mainTerms: main terms of the cost function, for example similarity measure or collision detection
        :type mainTerms: list of any regularization, collisionDetection, or similarityMeasure class
        :param regTerms: regularization terms of the cost function
        :type regTerms: list of any regularization, collisionDetection, or similarityMeasure class
        :return ptsDis: displaced points in spatial configuration
        :rtype ptsDis: torch.Tensor
        :return loss: accumulated loss of mainTerm and regTerms
        :rtype loss: torch.Tensor
        """
        ptsDis,loss=self.integrator(self.pts,velocityField,self.interpolator,mainTerms,regTerms)
        return ptsDis,loss
    def apply(self,pts: Tensor,velocityField: Tensor)->Tensor:
        """Displaces points with nonrigid displacement.
        :param velocityField: set of velocity fields (time frames, dimensions, # control points dim 1, # control points dim 2 [,# control points dim 3])
        :type velocityField: torch.Tensor
        :return ptsDis: displaced points as cartesian coordinates (# points, dim)
        :rtype ptsDis: torch.Tensor
        """
        ptsDis, loss = self.integrator(pts, velocityField, self.interpolator, None, None)
        return ptsDis

class affineTransformation():
    """ Class for affine displacement.
    :param pts: points in material configuration
    :type pts: torch.Tensor
    """
    def __init__(self,pts: Tensor):
        """Constructor method."""
        self.pts=pts
        self.dimension = pts.size(1)
        self.device = pts.device
    def __call__(self, affineMat: Tensor,mainTerms: Optional[list] = None,regTerms: Optional[list] = None)->Tuple[Tensor,Tensor]:
        """ Displaces points with an affine transformation. Evaluates mainTerm and regTerms on the fly.
        :param affineMat: affine transformation matrix; either 3x3 or 4x4;
        :type affineMat: torch.Tensor
        :param mainTerms: main terms of the cost function, for example similarity measure or collision detection
        :type mainTerms: list of any regularization, collisionDetection, or similarityMeasure class
        :param regTerms: regularization terms of the cost function
        :type regTerms: list of any regularization, collisionDetection, or similarityMeasure class
        :return ptsDis: displaced points in spatial configuration
        :rtype ptsDis: torch.Tensor
        :return loss: accumulated loss of mainTerms and regTerms
        :rtype loss: torch.Tensor
        """
        ptsDis = torch.cat((self.pts, torch.ones((self.pts.shape[0], 1), device=self.device, dtype=self.pts.dtype)), 1)
        ptsDis = affineMat.mm(ptsDis.t()).t()[:, :self.dimension]
        loss=0
        if mainTerms is not None:
            for term in mainTerms:
                loss=loss+term(dis=ptsDis-self.pts)
        elif regTerms is not None:
            for term in regTerms:
                loss=loss+term(dis=ptsDis-self.pts)
        else:
            loss=None
        return ptsDis,loss
    def apply(self,pts: Tensor,affineMat: Tensor)-> Tensor:
        """ Displaces points with an affine transformation.
        :param affineMat: affine transformation matrix; either 3x3 or 4x4;
        :type affineMat: torch.Tensor
        :return ptsDis: displaced points in spatial configuration
        :rtype ptsDis: torch.Tensor
        """
        ptsDis = torch.cat((pts, torch.ones((pts.shape[0], 1), device=self.device, dtype=pts.dtype)), 1)
        ptsDis = affineMat.mm(ptsDis.t()).t()[:, :self.dimension]
        return ptsDis







import torch
import numpy as np
import os
from .. import interpolation, geometricTransformations,  numericalIntegration
import re
import matplotlib.pyplot as plt
import pickle
from .postprocessing_utils import *
from typing import Optional, Type, Union, Tuple
from torch import Tensor

def visual_grid(device: str,workingDirectory: str,voxelToMm: Optional[list]=None,segmentsOfInterest: Optional[list]=None,dimension: int=None,slice: int=None):
    """ Plots the deformation grid.
        :param device: sets the computation device, see torch
        :type device: string
        :param workingDirectory: path to working directory, see docs
        :type workingDirectory: string
        :param segmentsOfInterest: segmentations of interest list
        :type segmentsOfInterest: list
        :param voxelSizes: cell dimensions in mm
        :type voxelSizes: torch.Tensor
        :param dimension: if 3D registration, set with dimension should be displayed
        :type dimension: int
        :param slice: if 3D registration, set with slice in dimension should be displayed
        :type slice: int
    """

    #BASICS: load images
    movingImageMask=torch.unsqueeze(torch.from_numpy(np.load(os.path.join(workingDirectory, "moving_mask.npy"))), dim=0).to(device=device)
    movingImage=torch.unsqueeze(torch.from_numpy(np.load(os.path.join(workingDirectory, "moving.npy"))), dim=0).to(device=device)


    indices = np.indices(movingImage.cpu()[0].size())
    pts = np.empty((np.prod(movingImage.cpu().size()), len(movingImage[0].cpu().size())))
    for i, slide in enumerate(indices):
        pts[:, i] = slide.flatten()
    pts=torch.from_numpy(pts).to(device=device).float()
    pts_orig=pts.clone()

    if segmentsOfInterest is not None:
        inter = interpolation.nearest(device,
                                      scale=torch.ones(pts.size(1), device=device))
        ptsSegmentation = inter(pts, movingImageMask).flatten().long()

        mask = torch.zeros_like(ptsSegmentation, dtype=torch.bool)
        for segment in segmentsOfInterest:
            mask |= (ptsSegmentation == segment)
        pts=pts[mask]
        ptsSegmentation=ptsSegmentation[mask]
        pts=checkAffine(device,workingDirectory,pts,ptsSegmentation)
        pts=checkNonrigid(device,workingDirectory,pts,ptsSegmentation)
        p=pts_orig.clone()
        p[mask]=pts
        pts=p
    else:
        pts = checkAffine(device, workingDirectory, pts)
        pts = checkNonrigid(device, workingDirectory, pts)


    if dimension is not None and slice is not None:
        fig, ax = plt.subplots(1)
        if dimension==0:
            movingDIM = movingImage[0, slice, :, :]
            selector=pts_orig[:,dimension]==slice
            ptsTemp=pts[selector]
            pts_origTemp=pts_orig[selector]
            extent = [0, movingDIM.size(1) * voxelToMm[2].cpu(), 0, movingDIM.size(0)  * voxelToMm[1].cpu()]
            ax.imshow(movingDIM.cpu(), extent=extent, origin='lower')
            for x in range(movingImage.size(2)):
                ptsX=ptsTemp[pts_origTemp[:,1]==x]
                ax.plot(ptsX[:,2].cpu()*voxelToMm[2].cpu(),ptsX[:,1].cpu()*voxelToMm[1].cpu(),c='r')
            for y in range(movingImage.size(3)):
                ptsY=ptsTemp[pts_origTemp[:,2]==y]
                ax.plot(ptsY[:,2].cpu()*voxelToMm[2].cpu(),ptsY[:,1].cpu()*voxelToMm[1].cpu(),c='r')
            plt.show()

        elif dimension==1:
            movingDIM = movingImage[0, :, slice, :]
            selector = pts_orig[:, dimension] == slice
            ptsTemp = pts[selector]
            pts_origTemp = pts_orig[selector]
            extent = [0, movingDIM.size(1) * voxelToMm[2].cpu(), 0, movingDIM.size(0) * voxelToMm[0].cpu()]
            ax.imshow(movingDIM.cpu(), extent=extent, origin='lower')
            for x in range(movingImage.size(1)):
                ptsX = ptsTemp[pts_origTemp[:, 0] == x ]
                ax.plot(ptsX[:, 2].cpu()*voxelToMm[2].cpu(), ptsX[:, 0].cpu()*voxelToMm[0].cpu(), c='r')
            for y in range(movingImage.size(3)):
                ptsY = ptsTemp[pts_origTemp[:, 2] == y]
                ax.plot(ptsY[:, 2].cpu()*voxelToMm[2].cpu(), ptsY[:, 0].cpu()*voxelToMm[0].cpu(), c='r')
            plt.show()
        elif dimension==2:
            movingDIM = movingImage[0, :, :, slice]

            selector = pts_orig[:, dimension] == slice
            ptsTemp = pts[selector]
            pts_origTemp = pts_orig[selector]
            extent = [0, movingDIM.size(1) * voxelToMm[1].cpu(), 0, movingDIM.size(0) * voxelToMm[0].cpu()]
            ax.imshow(movingDIM.cpu(), extent=extent, origin='lower')

            for x in range(movingImage.size(1)):
                ptsX = ptsTemp[pts_origTemp[:, 0] == x]
                ax.plot(ptsX[:, 1].cpu()*voxelToMm[1].cpu(), ptsX[:, 0].cpu()*voxelToMm[0].cpu(), c='r')

            for y in range(movingImage.size(2)):
                ptsY = ptsTemp[pts_origTemp[:, 1] == y]
                ax.plot(ptsY[:, 1].cpu()*voxelToMm[1].cpu(), ptsY[:, 0].cpu()*voxelToMm[0].cpu(), c='r')

            plt.show()
        else:
            print("check dimension")
    else:
        fig, ax = plt.subplots(1)

        movingDIM = movingImage[0]

        extent = [0, movingDIM.size(1) * voxelToMm[1].cpu(), 0, movingDIM.size(0) * voxelToMm[0].cpu()]
        for x in range(movingImage.size(1)):
            ptsX = pts[pts_orig[:, 0] == x]
            ax.plot(ptsX[:, 1].cpu()*voxelToMm[1].cpu(), ptsX[:, 0].cpu()*voxelToMm[0].cpu(), c='r')
        for y in range(movingImage.size(2)):
            ptsY = pts[pts_orig[:, 1] == y]
            ax.plot(ptsY[:, 1].cpu()*voxelToMm[1].cpu(), ptsY[:, 0].cpu()*voxelToMm[0].cpu(), c='r')
        plt.xticks([])  # Removes the x-axis numbers
        plt.yticks([])
        plt.show()






import torch
from typing import Optional, Type, Union, Tuple
from torch import Tensor

class linOp():
    """ Class for linear operator regularization of velocity fields
    :param coef: weight of linOp
    :type coef: float
    :param lam: coefficient for laplacian term
    :type lam: float
    :param mu: coefficient for identity term
    :type mu: float
    :param pointsMask: a mask for pts if only a subset of pts needs to be regularized
    :type pointsMask: torch.Tensor
    :param pointsMaskLabel: the mask label for pointsMask that needs to be regularized
    :type pointsMaskLabel: int
    """
    def __init__(self,coef: float=1,mu: float=1,lam: float=1,pointsMask: Optional[Tensor] = None,pointsMaskLabel: Optional[Tensor] = None):
        """constructor method"""
        self.mu=mu
        self.lam=lam
        self.coef=coef
        self.pointsMask=pointsMask
        self.pointsMaskLabel=pointsMaskLabel

    def __call__(self,vel: Tensor=0,vel_lap: Tensor=0,**kwargs)->Tensor:
        """Calculates the linear operator.
            :param vel: points velocities
            :type vel: torch.Tensor
            :param vel_lap: laplacian of point velocities
            :type vel_lap: torch.Tensor
            :return: linear operator norm
            :rtype: torch.Tensor"""
        if self.pointsMask is not None:
            vel = vel[self.pointsMask == self.pointsMaskLabel]
            vel_lap = vel_lap[self.pointsMask == self.pointsMaskLabel]
        else:
            pass
        return self.coef*torch.sum(torch.norm(self.mu*vel_lap+self.lam*vel.t(),dim=1))





import torch
import numpy as np
from scipy import ndimage
import os
from .. import  geometricTransformations, interpolation
import pickle
from .postprocessing_utils import *
import math
from typing import Optional, Type, Union, Tuple
from torch import Tensor


def measure_jacobian(device: str,workingDirectory: str,voxelToMm: Optional[Tensor]=None,segmentsOfInterest: Optional[list]=None)->dict:
    """ Calculates the jacobian of teh deformation field.
        :param device: sets the computation device, see torch
        :type device: string
        :param workingDirectory: path to working directory, see docs
        :type workingDirectory: string
        :param segmentsOfInterest: segmentations of interest list
        :type segmentsOfInterest: list
        :param voxelSizes: cell dimensions in mm
        :type voxelSizes: torch.Tensor
        :return fields: dictionary of deformation field jacobians
        :rtype fields: dict
    """
    movingImageMask=torch.unsqueeze(torch.from_numpy(np.load(os.path.join(workingDirectory, "moving_mask.npy"))), dim=0).to(device=device)
    fixedImageMask=torch.unsqueeze(torch.from_numpy(np.load(os.path.join(workingDirectory, "fixed_mask.npy"))), dim=0).to(device=device)
    indices = np.indices(movingImageMask.cpu()[0].size())
    pts = np.empty((np.prod(movingImageMask.cpu().size()), len(movingImageMask[0].cpu().size())))
    for i, slide in enumerate(indices):
        pts[:, i] = slide.flatten()
    pts=torch.from_numpy(pts).to(device=device).float()
    fields={}
    if pts.size(1)==3:
        if segmentsOfInterest is not None:
            for segment in segmentsOfInterest:
                ptsSegmentation = torch.ones(pts.size(0), dtype=torch.bool)*segment
                pts=checkAffine(device,workingDirectory,pts,ptsSegmentation)
                pts=checkNonrigid(device,workingDirectory,pts,ptsSegmentation)

                phiX = (pts[:, 0]).flatten().reshape(fixedImageMask.size())[0] * voxelToMm[0]
                phiY = (pts[:, 1]).flatten().reshape(fixedImageMask.size())[0] * voxelToMm[1]
                phiZ = (pts[:, 2]).flatten().reshape(fixedImageMask.size())[0] * voxelToMm[2]

                xphiX, yphiX, zphiX = compute_gradient_central_diff3D(phiX)
                xphiY, yphiY, zphiY = compute_gradient_central_diff3D(phiY)
                xphiZ, yphiZ, zphiZ = compute_gradient_central_diff3D(phiZ)

                id = torch.eye(3)
                id = id.reshape((1, 3, 3))
                id = id.repeat(phiX.size(0), phiX.size(1), phiX.size(2), 1, 1)

                jacPhi = torch.zeros(id.size())
                jacPhi[:, :, :, 0, 0] = xphiX
                jacPhi[:, :, :, 0, 1] = yphiX
                jacPhi[:, :, :, 0, 2] = zphiX
                jacPhi[:, :, :, 1, 0] = xphiY
                jacPhi[:, :, :, 1, 1] = yphiY
                jacPhi[:, :, :, 1, 2] = zphiY
                jacPhi[:, :, :, 2, 0] = xphiZ
                jacPhi[:, :, :, 2, 1] = yphiZ
                jacPhi[:, :, :, 2, 2] = zphiZ
                fields[str(segment)]=torch.linalg.det(jacPhi).cpu().numpy().tolist()

        else:
            pts=checkAffine(device,workingDirectory,pts)
            pts=checkNonrigid(device,workingDirectory,pts)

            phiX = (pts[:, 0]).flatten().reshape(fixedImageMask.size())[0] * voxelToMm[0]
            phiY = (pts[:, 1]).flatten().reshape(fixedImageMask.size())[0] * voxelToMm[1]
            phiZ = (pts[:, 2]).flatten().reshape(fixedImageMask.size())[0] * voxelToMm[2]

            xphiX, yphiX, zphiX = compute_gradient_central_diff3D(phiX)
            xphiY, yphiY, zphiY = compute_gradient_central_diff3D(phiY)
            xphiZ, yphiZ, zphiZ = compute_gradient_central_diff3D(phiZ)

            id = torch.eye(3)
            id = id.reshape((1, 3, 3))
            id = id.repeat(phiX.size(0), phiX.size(1), phiX.size(2), 1, 1)

            jacPhi = torch.zeros(id.size())
            jacPhi[:, :, :, 0, 0] = xphiX
            jacPhi[:, :, :, 0, 1] = yphiX
            jacPhi[:, :, :, 0, 2] = zphiX
            jacPhi[:, :, :, 1, 0] = xphiY
            jacPhi[:, :, :, 1, 1] = yphiY
            jacPhi[:, :, :, 1, 2] = zphiY
            jacPhi[:, :, :, 2, 0] = xphiZ
            jacPhi[:, :, :, 2, 1] = yphiZ
            jacPhi[:, :, :, 2, 2] = zphiZ
            fields["overall"] = torch.linalg.det(jacPhi).cpu().numpy().tolist()

    else:
        if segmentsOfInterest is not None:
            for segment in segmentsOfInterest:
                ptsSegmentation = torch.ones(pts.size(0), dtype=torch.bool) * segment
                pts = checkAffine(device, workingDirectory, pts, ptsSegmentation)
                pts = checkNonrigid(device, workingDirectory, pts, ptsSegmentation)

                phiX = (pts[:, 0]).flatten().reshape(fixedImageMask.size())[0] * voxelToMm[0]
                phiY = (pts[:, 1]).flatten().reshape(fixedImageMask.size())[0] * voxelToMm[1]

                xphiX, yphiX = compute_gradient_central_diff2D(phiX)
                xphiY, yphiY = compute_gradient_central_diff2D(phiY)

                id = torch.eye(2)
                id = id.reshape((1, 2, 2))
                id = id.repeat(phiX.size(0), phiX.size(1), 1, 1)

                jacPhi = torch.zeros(id.size())
                jacPhi[:, :, 0, 0] = xphiX
                jacPhi[:, :, 0, 1] = yphiX
                jacPhi[:, :, 1, 0] = xphiY
                jacPhi[:, :, 1, 1] = yphiY
                fields[str(segment)] = torch.linalg.det(jacPhi).cpu().numpy().tolist()

        else:
            pts = checkAffine(device, workingDirectory, pts)
            pts = checkNonrigid(device, workingDirectory, pts)

            phiX = (pts[:, 0]).flatten().reshape(fixedImageMask.size())[0] * voxelToMm[0]
            phiY = (pts[:, 1]).flatten().reshape(fixedImageMask.size())[0] * voxelToMm[1]

            xphiX, yphiX = compute_gradient_central_diff2D(phiX)
            xphiY, yphiY = compute_gradient_central_diff2D(phiY)

            id = torch.eye(2)
            id = id.reshape((1, 2, 2))
            id = id.repeat(phiX.size(0), phiX.size(1), 1, 1)

            jacPhi = torch.zeros(id.size())
            jacPhi[:, :, 0, 0] = xphiX
            jacPhi[:, :, 0, 1] = yphiX
            jacPhi[:, :, 1, 0] = xphiY
            jacPhi[:, :, 1, 1] = yphiY
            fields["overall"] = torch.linalg.det(jacPhi).cpu().numpy().tolist()
    return fields














import warnings

import torch
import numpy as np
import os
from .. import interpolation, geometricTransformations,  numericalIntegration
import re
import matplotlib.pyplot as plt
import pickle
from .postprocessing_utils import *
from typing import Optional, Type, Union, Tuple
from torch import Tensor

def measure_landmarks(device: str,workingDirectory: str,voxelToMm: Optional[list]=None,segmentsOfInterest: Optional[list]=None)->Tuple[Tensor,Tensor]:
    """ Calculates the target registration error.
        :param device: sets the computation device, see torch
        :type device: string
        :param workingDirectory: path to working directory, see docs
        :type workingDirectory: string
        :param segmentsOfInterest: segmentations of interest list
        :type segmentsOfInterest: list
        :param voxelSizes: cell dimensions in mm
        :type voxelSizes: torch.Tensor
        :return : mean and standard deviation of TRE
        :rtype : Tensor,Tensor
    """
    # BASICS: load images
    movingImageMask = torch.unsqueeze(torch.from_numpy(np.load(os.path.join(workingDirectory, "moving_mask.npy"))),
                                      dim=0).to(device=device)
    landmarkCoordinatesStart = torch.from_numpy(np.load(os.path.join(workingDirectory, "moving_landmarks.npy"))).to(
        device=device).float()
    landmarkCoordinatesEnd = torch.from_numpy(np.load(os.path.join(workingDirectory, "fixed_landmarks.npy"))).to(
        device=device).float()

    if segmentsOfInterest is not None:
        inter = interpolation.nearest(device,
                                      scale=torch.ones(landmarkCoordinatesStart.size(1), device=device))
        lmSegmentation = inter(landmarkCoordinatesStart, movingImageMask).flatten().long()

        mask = torch.zeros_like(lmSegmentation, dtype=torch.bool)
        for segment in segmentsOfInterest:
            mask |= (lmSegmentation == segment)
        landmarkCoordinatesStart=landmarkCoordinatesStart[mask]
        lmSegmentation=lmSegmentation[mask]
        landmarkCoordinatesEnd=landmarkCoordinatesEnd[mask]
        landmarkCoordinatesStart=checkAffine(device,workingDirectory,landmarkCoordinatesStart,lmSegmentation)
        landmarkCoordinatesStart=checkNonrigid(device,workingDirectory,landmarkCoordinatesStart,lmSegmentation)
    else:
        landmarkCoordinatesStart=checkAffine(device,workingDirectory,landmarkCoordinatesStart)
        landmarkCoordinatesStart=checkNonrigid(device,workingDirectory,landmarkCoordinatesStart)

    diff = landmarkCoordinatesStart - landmarkCoordinatesEnd
    diff = diff * voxelToMm
    dist = torch.norm(diff, dim=1)
    print("Mean TRE: ", torch.mean(dist), "STD TRE: ", torch.std(dist))
    diff = torch.round(landmarkCoordinatesStart) - landmarkCoordinatesEnd
    diff = diff * voxelToMm
    dist = torch.norm(diff, dim=1)
    print("fit to closest pixel value approach")
    print("Mean TRE: ", torch.mean(dist), "STD TRE: ", torch.std(dist))

    return torch.mean(dist), torch.std(dist), dist







import torch
import tqdm
from typing import Optional, Type, Union, Tuple
from torch import Tensor
from ..geometricTransformations import *
from .optim_gradientDescentBacktracking import *

class constrainerADMM:
    """Template class for ADMM coupling constraints.
                :param c: Tensor with additive constant
                :type c: torch.Tensor
                :param dualVariable: Tensor with dual variable
                :type dualVariable: torch.Tensor
                :param decisionVariablesXCoef: coefficient for X decision variable
                :type decisionVariablesXCoef: float
                :param decisionVariablesZCoef: coefficient for X decision variable
                :type decisionVariablesZCoef: float"""
    def __init__(self,c: Tensor,dualVariable: Tensor,decisionVariablesXCoef:float,decisionVariablesZCoef:float):
        """Constructor method."""
        self.c=c
        self.dualVariable=dualVariable
        self.decisionVariablesXCoef=decisionVariablesXCoef
        self.decisionVariablesZCoef=decisionVariablesZCoef
    def __call__(self,decisionVariable: Tensor,decisionVariableFixed: Tensor,decisionVariableCoef: float,decisionVariableFixedCoef: float):
        pass
    def updateDualVariable(self, decisionVariablesX: Tensor, decisionVariablesZ: Tensor):
        pass

class constrainerEulerianADMM(constrainerADMM):
    """Class for Eulerian ADMM coupling constraints. See Template class.
    """
    def __call__(self, decisionVariables: Tensor,decisionVariablesCoef: float,fixedDecisionVariables: Tensor,fixedDecisionVariablesCoef: float)->Tensor:
        """Calculates constrainer loss.
                :param decisionVariables: Tensor with decision variables
                :type decisionVariables: torch.Tensor
                :param decisionVariablesCoef: coefficient for decision variables
                :type decisionVariablesCoef: float
                :param fixedDecisionVariables: Tensor with fixed decision variables
                :type fixedDecisionVariables: torch.Tensor
                :param fixedDecisionVariablesCoef: coefficient for fixed decision variables
                :type fixedDecisionVariablesCoef: float
                :return : constrainer loss
                :rtype : torch.Tensor
        """
        return decisionVariablesCoef * decisionVariables + fixedDecisionVariablesCoef * fixedDecisionVariables - self.c +self.dualVariable
    def updateDualVariable(self, decisionVariablesX: Tensor, decisionVariablesZ: Tensor):
        """Updates the dual variable.
                :param decisionVariablesX: Tensor with decision variables
                :type decisionVariablesX: torch.Tensor
                :param decisionVariablesZ: Tensor with decision variables
                :type decisionVariablesZ: torch.Tensor
        """
        self.dualVariable = self.dualVariable + self.decisionVariablesXCoef * decisionVariablesX.data + self.decisionVariablesZCoef * decisionVariablesZ.data - self.c
class constrainerLagrangianADMM(constrainerADMM):
    """Class for Lagrangian ADMM coupling constraints. See Template class.
        :param transformer: transformer
        :type transformer: affineTransformation or nonrigidDeformation class
        :param pts: coupling points
        :type pts: torch.Tensor
    """
    def __init__(self,c: Tensor,dualVariable: Tensor,decisionVariablesXCoef: float,decisionVariablesZCoef: float,transformer: Union[affineTransformation,nonrigidDeformation],pts: Tensor):
        super().__init__(c,dualVariable,decisionVariablesXCoef,decisionVariablesZCoef)
        self.transformer=transformer
        self.pts=pts
    def __call__(self, decisionVariables: Tensor,decisionVariablesCoef: float,fixedDecisionVariables: Tensor,fixedDecisionVariablesCoef: float)->Tensor:
        """Calculates constrainer loss.
                :param decisionVariables: Tensor with decision variables
                :type decisionVariables: torch.Tensor
                :param decisionVariablesCoef: coefficient for decision variables
                :type decisionVariablesCoef: float
                :param fixedDecisionVariables: Tensor with fixed decision variables
                :type fixedDecisionVariables: torch.Tensor
                :param fixedDecisionVariablesCoef: coefficient for fixed decision variables
                :type fixedDecisionVariablesCoef: float
                :return : constrainer loss
                :rtype : torch.Tensor
        """
        pts=self.transformer.apply(self.pts,decisionVariables)
        ptsFixed=self.transformer.apply(self.pts,fixedDecisionVariables)
        return decisionVariablesCoef * pts + fixedDecisionVariablesCoef * ptsFixed - self.c +self.dualVariable
    def updateDualVariable(self, decisionVariablesX: Tensor, decisionVariablesZ: Tensor):
        """Updates the dual variable.
                :param decisionVariablesX: Tensor with decision variables
                :type decisionVariablesX: torch.Tensor
                :param decisionVariablesZ: Tensor with decision variables
                :type decisionVariablesZ: torch.Tensor
        """
        ptsX=self.transformer.apply(self.pts,decisionVariablesX.data)
        ptsZ=self.transformer.apply(self.pts,decisionVariablesZ.data)
        self.dualVariable = self.dualVariable + self.decisionVariablesXCoef * ptsX + self.decisionVariablesZCoef * ptsZ - self.c

class closureADMM():
    """ Class for a custom closure function (see pytorch docs) to evaluate the loss function, specifically for ADMM algorithm.
            :param optimizer: optimizer class used for the minimization problem
            :type optimizer: torch.optim.optimizer class or custom class
            :param decisionVariablesCoef: coefficient for decisionVariables
            :type decisionVariablesCoef: int, float ot torch.Tensor
            :param decisionVariablesFixedCoef: coefficient for fixed decisionVariables
            :type decisionVariablesFixedCoef: int, float ot torch.Tensor
            :param rho: penalty parameter to weigh the trade off,see ADMM
            :type rho: int or float
            :param mainTerms: main term to be minimized
            :type mainTerms: simMeasure or regularizer class
            :param mainTermCoef: coefficient for mainTerm
            :type mainTermCoef: either int or float
            :param regTerms: list of regularizers to constrain the minimization problem
            :type regTerms: list of regularizer classes
            :param regTermsCoefs: list of coefficients for the regularizers
            :type regTermsCoefs: list of ints or floats
            :return: backpropagated accumulated loss
            :rtype: torch.Tensor
            """
    def __init__(self,optimizer: Union[gradientDescentBacktracking,torch.optim.Optimizer],transformer: Union[affineTransformation,nonrigidDeformation],rho: Optional[float]=1.0,mainTerms: Optional[list]=None, regTerms: Optional[list]=[]):
        """Constructor method.            """
        self.optimizer=optimizer
        self.rho=rho
        self.mainTerms=mainTerms
        self.regTerms=regTerms
        self.transformer=transformer

    def __call__(self,decisionVariables: Tensor,decisionVariablesCoef: float,fixedDecisionVariables: Tensor,fixedDecisionVariablesCoef: float,constrainer: Union[Type[constrainerEulerianADMM],Type[constrainerLagrangianADMM]])->Tensor:
        """ Initiates the forward pass for ADMM. In case of gradients switched on, also zeros the gradients first.
                :param decisionVariables: tensor with decision variables
                :type decisionVariables: torch.Tensor
                :param decisionVariablesCoef: coefficients for decision variables
                :type decisionVariablesCoef: float
                :param fixedDecisionVariables: tensor with fixed decision variables
                :type fixedDecisionVariables: torch.Tensor
                :param fixedDecisionVariablesCoef: coefficients for decision variables
                :type fixedDecisionVariablesCoef: float
                :param constrainer: measures the deviation from constraints, see ADMM
                :type constrainer: constrainerEulerianADMM or constrainerLagrangianADMM class
                :return: backpropagated accumulated loss
                :rtype: torch.Tensor
                """
        if decisionVariables.requires_grad:
            self.optimizer.zero_grad()
            _,loss=self.transformer(decisionVariables,self.mainTerms,self.regTerms)
            loss=loss+0.5*self.rho*(torch.norm(constrainer(decisionVariables,decisionVariablesCoef,fixedDecisionVariables,fixedDecisionVariablesCoef))**2)
            loss.backward()
        else:
            _, loss = self.transformer(decisionVariables, self.mainTerms, self.regTerms)
            loss=loss+0.5*self.rho*(torch.norm(constrainer(decisionVariables,decisionVariablesCoef,fixedDecisionVariables,fixedDecisionVariablesCoef))**2)
        return loss
def algorithmADMM(iterations: int,subsolverIterations: int,constrainer: Union[constrainerEulerianADMM,constrainerLagrangianADMM],optimizerF: Union[gradientDescentBacktracking,torch.optim.Optimizer],optimizerG: Union[gradientDescentBacktracking,torch.optim.Optimizer],closureF: Type,closureG: Type,decisionVariablesX: Tensor,decisionVariablesZ: Tensor,rho: float)->dict:
    """ Starts the ADMM iterative optimization scheme.
            :param iterations: number of steps in scheme
            :type iterations: int
            :param subsolverIterations: number of steps for every subsolver
            :type subsolverIterations: int
            :param constrainer: measures the deviation from constraints, see ADMM
            :type constrainer: constrainerEulerianADMM or constrainerLagrangianADMM class
            :param optimizerF: optimizer class used for the first minimization problem
            :type optimizerF: gradientDescentBacktracking or torch.optim.Optimizer class
            :param optimizerG: optimizer class used for the second minimization problem
            :type optimizerG: gradientDescentBacktracking or torch.optim.Optimizer class
            :param closureF: closure function to calculate loss and backpropagate for the first minimization problem.
            :type closureF: either a function or a class with a __call__ method
            :param closureG: closure function to calculate loss and backpropagate for the second minimization problem.
            :type closureG: either a function or a class with a __call__ method
            :param decisionVariablesX: Tensor with decision variables for the first problem
            :type decisionVariablesX: torch.Tensor
            :param decisionVariablesZ: Tensor with decision variables for the second problem
            :type decisionVariablesZ: torch.Tensor'
            :param rho: penalty parameter
            :type rho: float
            :return dict: dictionary with gradient history and objective function history
            :rtype dict: dict
            """
    objectiveLossFHistory=[]
    objectiveLossGHistory=[]
    primalResidualHistory=[]
    dualResidualHistory=[]
    dualVariableHistory=[]
    for i in tqdm.tqdm(range(iterations),desc="Progress"):
        z_old=decisionVariablesZ.data.clone()
        for j in range(subsolverIterations):
            lossX=optimizerF.step(lambda: closureF(decisionVariablesX,constrainer.decisionVariablesXCoef,decisionVariablesZ.data,constrainer.decisionVariablesZCoef,constrainer))
        objectiveLossFHistory.append(lossX.cpu().item())
        for j in range(subsolverIterations):
            lossZ=optimizerG.step(lambda: closureG(decisionVariablesZ,constrainer.decisionVariablesZCoef,decisionVariablesX.data,constrainer.decisionVariablesXCoef,constrainer))
        constrainer.updateDualVariable(decisionVariablesX,decisionVariablesZ)
        objectiveLossGHistory.append(lossZ.cpu().item())
        primalResidualHistory.append(torch.norm((decisionVariablesX.data - decisionVariablesZ.data)).cpu().item())
        dualResidualHistory.append(torch.norm(-rho * (decisionVariablesZ.data - z_old)).cpu().item())
        dualVariableHistory.append(torch.norm(constrainer.dualVariable).cpu().item())
    dict = {"objectiveLossFHistory": objectiveLossFHistory,
            "objectiveLossGHistory": objectiveLossGHistory,
            "primalResidualHistory": primalResidualHistory,
            "dualResidualHistory": dualResidualHistory,
            "dualVariableHistory": dualVariableHistory
            }
    return dict



def algorithmADMMStochastic(device: str,iterations: int,subsolverIterations: int,constrainer: Union[constrainerEulerianADMM,constrainerLagrangianADMM],optimizerF: Union[gradientDescentBacktracking,torch.optim.Optimizer],optimizerG: Union[gradientDescentBacktracking,torch.optim.Optimizer],closureF: Type,closureG: Type,decisionVariablesX: Tensor,decisionVariablesZ: Tensor,rho: float,evalPoints1: Tensor,evalPointsIntensities1: Tensor,percentage1: float,stochasticTerms1: list,evalPoints2: Tensor,evalPointsIntensities2: Tensor,percentage2: float,stochasticTerms2: list,pointsMaskLabel1: Optional[Tensor]=None,pointsMaskLabel2: Optional[Tensor]=None)->dict:
    """ Starts the ADMM iterative optimization scheme.  It uses stochastic selection of evaluation points.
            :param device: computation device, see torch docs
            :type device: str
            :param iterations: number of steps in scheme
            :type iterations: int
            :param subsolverIterations: number of steps for every subsolver
            :type subsolverIterations: int
            :param constrainer: measures the deviation from constraints, see ADMM
            :type constrainer: constrainerEulerianADMM or constrainerLagrangianADMM class
            :param optimizerF: optimizer class used for the first minimization problem
            :type optimizerF: gradientDescentBacktracking or torch.optim.Optimizer class
            :param optimizerG: optimizer class used for the second minimization problem
            :type optimizerG: gradientDescentBacktracking or torch.optim.Optimizer class
            :param closureF: closure function to calculate loss and backpropagate for the first minimization problem.
            :type closureF: either a function or a class with a __call__ method
            :param closureG: closure function to calculate loss and backpropagate for the second minimization problem.
            :type closureG: either a function or a class with a __call__ method
            :param decisionVariablesX: Tensor with decision variables for the first problem
            :type decisionVariablesX: torch.Tensor
            :param decisionVariablesZ: Tensor with decision variables for the second problem
            :type decisionVariablesZ: torch.Tensor'
            :param rho: penalty parameter
            :type rho: float
            :param evalPoints1: range of evaluation points available for selection for the first problem
            :type evalPoints1: torch.Tensor
            :param evalPointsIntensities1: range of intensities of evaluation points available for selection for the first problem
            :type evalPointsIntensities1: torch.Tensor
            :param percentage1: percentage of points used in each iteration ranging from [0;1] for the first problem
            :type percentage1: float
            :param stochasticTerms1: list of mainTerms and regTerms that need updates to their points in every iteration for the first problem
            :type stochasticTerms1: list
            :param pointsMaskLabel1: points mask labels for points that need to be exchanged for the first problem
            :type pointsMaskLabel1: torch.Tensor
            :param evalPoints2: range of evaluation points available for selection for the second problem
            :type evalPoints2: torch.Tensor
            :param evalPointsIntensities2: range of intensities of evaluation points available for selection for the second problem
            :type evalPointsIntensities2: torch.Tensor
            :param percentage2: percentage of points used in each iteration ranging from [0;1] for the second problem
            :type percentage2: float
            :param stochasticTerms2: list of mainTerms and regTerms that need updates to their points in every iteration for the second problem
            :type stochasticTerms2: list
            :param pointsMaskLabel2: points mask labels for points that need to be exchanged for the second problem
            :type pointsMaskLabel2: torch.Tensor
            :return dict: dictionary with gradient history and objective function history
            :rtype dict: dict
            """
    objectiveLossFHistory=[]
    objectiveLossGHistory=[]
    primalResidualHistory=[]
    dualResidualHistory=[]
    dualVariableHistory=[]
    if stochasticTerms1 is not None:
        length1=evalPoints1.size(0)
        numberOfPoints1=int(length1*percentage1)
    if stochasticTerms2 is not None:
        length2=evalPoints2.size(0)
        numberOfPoints2=int(length2*percentage2)
    for i in tqdm.tqdm(range(iterations),desc="Progress"):
        if stochasticTerms1 is not None:
            random_tensor1 = torch.randperm(evalPoints1.size(0))[:numberOfPoints1]
            evalPointsStochastic1 = evalPoints1[random_tensor1].to(device=device)
            evalPointsIntensitiesStochastic1 = evalPointsIntensities1[random_tensor1].to(device=device)
            for cnt,term in enumerate(stochasticTerms1):
                if hasattr(term, 'pts'):
                    if hasattr(term, 'pointsMask'):
                        if term.pointsMask is not None:
                            term.pts[term.pointsMask==pointsMaskLabel1[cnt]]=evalPointsStochastic1
                        else:
                            term.pts=evalPointsStochastic1
                    else:
                        term.pts=evalPointsStochastic1
                if hasattr(term, 'intensities'):
                    term.intensities=evalPointsIntensitiesStochastic1
        if stochasticTerms2 is not None:
            random_tensor2 = torch.randperm(evalPoints2.size(0))[:numberOfPoints2]
            evalPointsStochastic2 = evalPoints2[random_tensor2].to(device=device)
            evalPointsIntensitiesStochastic2 = evalPointsIntensities2[random_tensor2].to(device=device)
            for term in stochasticTerms2:
                if hasattr(term, 'pts'):
                    if hasattr(term, 'pts'):
                        if hasattr(term, 'pointsMask'):
                            if term.pointsMask is not None:
                                term.pts[term.pointsMask == pointsMaskLabel2[cnt]] = evalPointsStochastic2
                            else:
                                term.pts = evalPointsStochastic2
                        else:
                            term.pts = evalPointsStochastic2
                    if hasattr(term, 'intensities'):
                        term.intensities = evalPointsIntensitiesStochastic2
        z_old=decisionVariablesZ.data.clone()
        for j in range(subsolverIterations):
            lossX=optimizerF.step(lambda: closureF(decisionVariablesX,constrainer.decisionVariablesXCoef,decisionVariablesZ.data,constrainer.decisionVariablesZCoef,constrainer))

        objectiveLossFHistory.append(lossX.cpu().item())
        for j in range(subsolverIterations):
            lossZ=optimizerG.step(lambda: closureG(decisionVariablesZ,constrainer.decisionVariablesZCoef,decisionVariablesX.data,constrainer.decisionVariablesXCoef,constrainer))

        constrainer.updateDualVariable(decisionVariablesX,decisionVariablesZ)
        objectiveLossGHistory.append(lossZ.cpu().item())
        primalResidualHistory.append(torch.norm((decisionVariablesX.data - decisionVariablesZ.data)).cpu().item())
        dualResidualHistory.append(torch.norm(-rho * (decisionVariablesZ.data - z_old)).cpu().item())
        dualVariableHistory.append(torch.norm(constrainer.dualVariable).cpu().item())
    dict = {"objectiveLossFHistory": objectiveLossFHistory,
            "objectiveLossGHistory": objectiveLossGHistory,
            "primalResidualHistory": primalResidualHistory,
            "dualResidualHistory": dualResidualHistory,
            "dualVariableHistory": dualVariableHistory
            }
    return dict



import torch
from typing import Optional, Type, Union, Tuple
from torch import Tensor
from ..interpolation import *

class selfintersectionDetection2D:
    """See selfintersectionDetection docs."""
    def __init__(self,pts,sdf,interpolator,simplices,verticesUndeformed,coef,device,pointsMask,pointsMaskLabel,unmaskedDis,vertMask,vertMaskLabel,siPartner):
        """Constructor method."""
        self.sdf=sdf
        self.intP=interpolator
        self.simplices=simplices
        self.verticesUndeformed=verticesUndeformed
        self.device=device
        self.pts=pts
        self.coef=coef
        self.pointsMask=pointsMask
        self.pointsMaskLabel=pointsMaskLabel
        self.unmaskedDis=unmaskedDis
        self.vertMask=vertMask
        self.siPartner=siPartner
        self.vertMaskLabel=vertMaskLabel
    def __call__(self,dis: Tensor=None,**kwargs)->Tensor:
        """Executing 2D self collision detection.
        :param dis: displacement of pts
        :type dis: torch.Tensor
        :return: summed depth of selfintersecting points
        :rtype: torch.Tensor"""
        if self.pointsMask is not None:
            nodesMoved=self.pts[self.pointsMask==self.pointsMaskLabel].clone()
            dis=dis[self.pointsMask==self.pointsMaskLabel]
            if self.vertMask is not None:
                nodesMoved[self.vertMask==self.vertMaskLabel]=nodesMoved[self.vertMask==self.vertMaskLabel]+dis[self.vertMask==self.vertMaskLabel]
                nodesMoved[self.vertMask!=self.vertMaskLabel]=nodesMoved[self.vertMask!=self.vertMaskLabel]+self.unmaskedDis
                if self.siPartner is not None:
                    self.siPartner.unmaskedDis = dis[self.vertMask == self.vertMaskLabel].data
            else:
                nodesMoved=nodesMoved+dis
        else:
            nodesMoved=self.pts+dis
        l=self.loss2D(nodesMoved)*self.coef
        return l
    def barycentricToCartesian2D(self,barycentricCoords: Tensor, e1: Tensor,e2: Tensor,e3: Tensor)->Tensor:
        """Transformation from barycentric to cartesian coordinates in 2D.
            :param barycentricCoords: barycentric coordinates
            :type barycentricCoords: torch.Tensor
            :param e1,e2,e3: cartesian coordinates of simplex vertices
            :type e1,e2,e3: torch.Tensor
            :return p: cartesian coordinates
            :rtype p: torch.Tensor"""
        p=torch.unsqueeze(e1,dim=1)*barycentricCoords[:,0:2]+torch.unsqueeze(e2,dim=1)*barycentricCoords[:,2:4]+torch.unsqueeze(e3,dim=1)*barycentricCoords[:,4:]
        return p
    def vertexIdToCoordinates2D(self,intersectedSimplices: Tensor,vertexCoords: Tensor)->Tensor:
        """Get cartesian coordinates of vertices in 2D.
            :param intersectedSimplices: tensor of intersected Simplices
            :type intersectedSimplices: torch.Tensor
            :param vertexCoords: tensor of vertex coordinates
            :type vertexCoords: torch.Tensor
            :return: cartesian coordinates
            :rtype: torch.Tensor"""
        intersectedSimplices=intersectedSimplices[intersectedSimplices>=0]
        coordinates=torch.zeros((intersectedSimplices.size()[0],6),device=self.device)
        coordinates[:,0:2]=vertexCoords[self.simplices[intersectedSimplices.long()][:,0].long()]
        coordinates[:,2:4]=vertexCoords[self.simplices[intersectedSimplices.long()][:,1].long()]
        coordinates[:,4:]=vertexCoords[self.simplices[intersectedSimplices.long()][:,2].long()]
        return coordinates
    def cartesianToBarycentric2D(self,nodeCoordinates, p)->Tuple[Tensor,Tensor,Tensor]:
        """Transformation from cartesian to barycentric coordinates in 2D.
            :param nodeCoordinates: cartesian coordinates of simplex nodes
            :type nodeCoordinates: torch.Tensor
            :param p: cartesian coordiantes of point
            :type p: torch.Tensor
            :return e1,e2,e3: barycentric coordinates of point
            :rtype: torch.Tensors"""
        x1=nodeCoordinates[:,0]
        x2=nodeCoordinates[:,2]
        x3=nodeCoordinates[:,4]
        y1=nodeCoordinates[:,1]
        y2=nodeCoordinates[:,3]
        y3=nodeCoordinates[:,5]
        e1=((y2-y3)*(p[:,0]-x3)+(x3-x2)*(p[:,1]-y3))/((y2-y3)*(x1-x3)+(x3-x2)*(y1-y3))
        e2=((y3-y1)*(p[:,0]-x3)+(x1-x3)*(p[:,1]-y3))/((y2-y3)*(x1-x3)+(x3-x2)*(y1-y3))
        e3=1-e1-e2
        return e1,e2,e3
    def correctionForce2D(self,intersectingSimplices: Tensor,nodesDeformed: Tensor)->Tensor:
        """Calculates the depth of selfintersecting points to the nearest surface in 2D in the material configuration.
            :param nodesDeformed: displaced vertices
            :type nodesDeformed: torch.Tensor
            :param intersectingSimplices: tensor of the size of nodesDeformed where any non-negative entry indicates teh simplex the vertex is in; excludes the simplices the vertex was part of in the material configuration
            :type intersectingSimplices: torch.Tensor
            :return: squared and summed depth of selfintersecting points
            :rtype: torch.Tensor"""
        sel1=intersectingSimplices>=0
        sel2=self.vertMask == self.vertMaskLabel
        sel=sel1*sel2
        points = nodesDeformed[sel]
        intersectingSimplicesNew = intersectingSimplices[sel]
        coordinatesDeformed = self.vertexIdToCoordinates2D(intersectingSimplicesNew, nodesDeformed)
        coordinatesOrig = self.vertexIdToCoordinates2D(intersectingSimplicesNew, self.verticesUndeformed)
        e1, e2, e3 = self.cartesianToBarycentric2D(coordinatesDeformed, points)
        resamp = self.barycentricToCartesian2D(coordinatesOrig, e1, e2, e3)
        inter_sdf, _, _ = self.intP(resamp, self.sdf)
        return torch.sum(torch.sqrt(inter_sdf** 2))
    def intersectionDetection2D(self,nodesDeformed: Tensor, simplices: Tensor)->Tensor:
        """Computes whether a point lies inside a simplex after displacement. Points are assumed to be nodes of simplices, so hits with their own simplex are ignored. In 2D.
            :param nodesDeformed: displaced vertices
            :type nodesDeformed: torch.Tensor
            :param simplices: tensor of simplices containing node Ids
            :type simplices: torch.Tensor
            :return result: tensor of the size of nodesDeformed where any non-negative entry indicates teh simplex the vertex is in; excludes the simplices the vertex was part of in the material configuration
            :rtype result: torch.Tensor"""
        vertex0=nodesDeformed[simplices[:,0].long(),:]
        vertex1=nodesDeformed[simplices[:,1].long(),:]-vertex0
        vertex2=nodesDeformed[simplices[:,2].long(),:]-vertex0
        numberOfTriangles=len(simplices)
        vertice_1_expand=vertex1.T.reshape((2,1,numberOfTriangles))
        vertice_2_expand=vertex2.T.reshape((2,1,numberOfTriangles))
        mat = torch.cat((vertice_1_expand,vertice_2_expand), dim=1)
        inv_mat = torch.linalg.inv(mat.T).T
        if nodesDeformed.size(0)==2:
            nodesDeformed=nodesDeformed.reshape((1,2))
        numberNodesDeformed=nodesDeformed.shape[0]
        ori_expand=torch.repeat_interleave(vertex0[:,:,None], numberNodesDeformed, dim=2)
        new_points=torch.einsum('imk,kmj->kij',inv_mat,nodesDeformed.T-ori_expand)
        values=torch.all(new_points>=0, dim=1) & torch.all(new_points <=1, dim=1) & (torch.sum(new_points, dim=1)<=1)
        return_values=torch.nonzero(values)
        id_simplices=return_values[:,0]
        simplices=simplices[id_simplices]
        id_p=return_values[:,1]
        id_simplices_new=id_simplices[torch.any(simplices==torch.unsqueeze(id_p,dim=1),dim=1)==False]
        id_p_new=id_p[torch.any(simplices==torch.unsqueeze(id_p,dim=1),dim=1)==False]
        result = -torch.ones(numberNodesDeformed, dtype=id_simplices.dtype,device=self.device)
        result[id_p_new]=id_simplices_new
        return result
    def batching2D(self,nodesDeformed: Tensor)->Tensor:
        """ Computes which points are selfintersecting in batches in 2D.
            :param nodesDeformed: deformed mesh nodes
            :type nodesDeformed: torch.Tensor
            :return intersecting_points:  tensor of the size of nodesDeformed where any non-negative entry indicates teh simplex the vertex is in; excludes the simplices the vertex was part of in the material configuration
            :rtype intersecting_points: torch.Tensor"""
        numberTriangles=torch.tensor(self.simplices.size(0),device=self.device)
        max_num=1000
        iter=int((numberTriangles/max_num).floor()+1)
        results=torch.zeros((iter,nodesDeformed.size(0)),device=self.device)
        for cnt in range(iter):
            if cnt==0:
                simplicesTemp=self.simplices[cnt*max_num:(cnt+1)*max_num]
                resTemp=self.intersectionDetection2D(nodesDeformed, simplicesTemp)
                resTemp=torch.where(resTemp>=0,resTemp+(cnt*max_num),resTemp)
            elif cnt==iter-1:
                simplicesTemp=self.simplices[cnt*max_num:]
                resTemp=self.intersectionDetection2D(nodesDeformed, simplicesTemp)
                resTemp=torch.where(resTemp>=0,resTemp+(cnt*max_num),resTemp)
            else:
                simplicesTemp=self.simplices[cnt*max_num:(cnt+1)*max_num]
                resTemp=self.intersectionDetection2D(nodesDeformed, simplicesTemp)
                resTemp=torch.where(resTemp>=0,resTemp+(cnt*max_num),resTemp)
            results[cnt]=resTemp
        intersecting_points=torch.max(results,dim=0)[0]
        return intersecting_points
    def loss2D(self,nodesDeformed: Tensor)->Tensor:
        """Calculates the selfintersection loss in 2D.
            :param nodesDeformed: deformed mesh nodes
            :type nodesDeformed: torch.Tensor
            :return loss: summed depth of selfintersecting points
            :rtype loss: torch.Tensor"""
        intersectingNodes=self.batching2D(nodesDeformed)
        sel1=intersectingNodes>=0
        sel2=self.vertMask == self.vertMaskLabel
        sel=sel1*sel2
        if intersectingNodes[sel].size(0) == 0:
            loss=torch.zeros(1).to(device=self.device)
        else:
            loss=self.correctionForce2D(intersectingNodes,nodesDeformed)
        return loss
    def apply(self,nodes: Tensor, simplices: Tensor, coords: Tensor)->Tensor:
        """Checks if any coordinates in coords are covered by 2 or more simplices.
            :param nodes: vertices
            :type nodes: torch.Tensor
            :param simplices: tensor of simplices containing node Ids
            :type simplices: torch.Tensor
            :param coords: coordinates to be checked
            :type coords: torch.Tensor
            :return: tensor of coordinates covered by multiple simplices
            :rtype: torch.Tensor"""
        vertex1 = nodes[simplices[:, 0].long(), :]
        vertex2 = nodes[simplices[:, 1].long(), :] - vertex1
        vertex3 = nodes[simplices[:, 2].long(), :] - vertex1
        numberOfTetrahedras = len(simplices)
        vertex2_expand = vertex2.T.reshape((2, 1, numberOfTetrahedras))
        vertex3_expand = vertex3.T.reshape((2, 1, numberOfTetrahedras))
        mat = torch.cat((vertex2_expand, vertex3_expand), dim=1)
        inv_mat = torch.linalg.inv(mat.T).T  #
        if coords.size(0) == 3:
            coords = coords.reshape((1, 3))
        numberCoords = coords.shape[0]
        vertex1_expand = torch.repeat_interleave(vertex1[:, :, None], numberCoords, dim=2)
        new_points = torch.einsum('imk,kmj->kij', inv_mat, coords.T - vertex1_expand)
        values = torch.all(new_points >= 0, dim=1) & torch.all(new_points <= 1, dim=1) & (
                    torch.sum(new_points, dim=1) <= 1)
        return_values = torch.nonzero(values)
        flattened = return_values[:, 1].flatten()
        unique_values, counts = torch.unique(flattened, return_counts=True)
        duplicate_values = unique_values[counts > 1]
        return coords[duplicate_values, :]

class selfintersectionDetection3D():
    """See selfintersectionDetection docs."""
    def __init__(self,pts,sdf,interpolator,simplices,verticesUndeformed,coef,device,pointsMask,pointsMaskLabel,unmaskedDis,vertMask,vertMaskLabel,siPartner):
        """Constructor method."""
        self.sdf=sdf
        self.intP=interpolator
        self.simplices=simplices
        self.verticesUndeformed=verticesUndeformed
        self.device=device
        self.coef=coef
        self.pointsMask=pointsMask
        self.pointsMaskLabel=pointsMaskLabel
        self.unmaskedDis=unmaskedDis
        self.vertMask=vertMask
        self.siPartner=siPartner
        self.pts=pts
        self.vertMaskLabel=vertMaskLabel
    def __call__(self,dis: Tensor=None,**kwargs)->Tensor:
        """Executing 3D self collision detection.
        :param dis: displacement of pts
        :type dis: torch.Tensor
        :return: summed depth of selfintersecting points
        :rtype: torch.Tensor"""
        if self.pointsMask is not None:
            nodesMoved=self.pts[self.pointsMask==self.pointsMaskLabel].clone()
            dis=dis[self.pointsMask==self.pointsMaskLabel]
            if self.vertMask is not None:
                nodesMoved[self.vertMask==self.vertMaskLabel]=nodesMoved[self.vertMask==self.vertMaskLabel]+dis[self.vertMask==self.vertMaskLabel]
                nodesMoved[self.vertMask!=self.vertMaskLabel]=nodesMoved[self.vertMask!=self.vertMaskLabel]+self.unmaskedDis
                if self.siPartner is not None:
                    self.siPartner.unmaskedDis = dis[self.vertMask == self.vertMaskLabel].data
            else:
                nodesMoved=nodesMoved+dis

        else:
            nodesMoved=self.pts+dis
        l=self.loss3D(nodesMoved)*self.coef
        return l
    def intersectionDetection3D(self,nodesDeformed: Tensor, simplices: Tensor)->Tensor:
        """Computes whether a point lies inside a simplex after displacement. Points are assumed to be nodes of simplices, so hits with their own simplex are ignored. In 3D.
            :param nodesDeformed: displaced vertices
            :type nodesDeformed: torch.Tensor
            :param simplices: tensor of simplices containing node Ids
            :type simplices: torch.Tensor
            :return result: tensor of the size of nodesDeformed where any non-negative entry indicates teh simplex the vertex is in; excludes the simplices the vertex was part of in the material configuration
            :rtype result: torch.Tensor"""
        vertex1 = nodesDeformed[simplices[:, 0].long(), :]
        vertex2 = nodesDeformed[simplices[:, 1].long(), :] - vertex1
        vertex3 = nodesDeformed[simplices[:, 2].long(), :] - vertex1
        vertex4 = nodesDeformed[simplices[:, 3].long(), :] - vertex1
        numberOfTetrahedras = len(simplices)
        vertex2_expand = vertex2.T.reshape((3, 1, numberOfTetrahedras))
        vertex3_expand = vertex3.T.reshape((3, 1, numberOfTetrahedras))
        vertex4_expand = vertex4.T.reshape((3, 1, numberOfTetrahedras))
        mat = torch.cat((vertex2_expand, vertex3_expand, vertex4_expand), dim=1)
        inv_mat = torch.linalg.inv(mat.T).T  #
        if nodesDeformed.size(0) == 3:
            nodesDeformed = nodesDeformed.reshape((1, 3))
        numberNodesDeformed = nodesDeformed.shape[0]
        vertex1_expand = torch.repeat_interleave(vertex1[:, :, None], numberNodesDeformed, dim=2)
        new_points = torch.einsum('imk,kmj->kij', inv_mat, nodesDeformed.T - vertex1_expand)
        values = torch.all(new_points >= 0, dim=1) & torch.all(new_points <= 1, dim=1) & (torch.sum(new_points, dim=1) <= 1)
        return_values = torch.nonzero(values)
        id_tet = return_values[:, 0]
        simplices = simplices[id_tet]
        id_p = return_values[:, 1]
        id_tet_new = id_tet[torch.any(simplices == torch.unsqueeze(id_p, dim=1), dim=1) == False]
        id_p_new = id_p[torch.any(simplices == torch.unsqueeze(id_p, dim=1), dim=1) == False]
        res = -torch.ones(numberNodesDeformed, dtype=id_tet.dtype, device=self.device)  # Sentinel value
        res[id_p_new] = id_tet_new
        return res
    def batching3D(self,nodesDeformed: Tensor)->Tensor:
        """ Computes which points are selfintersecting in batches in 3D.
            :param nodesDeformed: deformed mesh nodes
            :type nodesDeformed: torch.Tensor
            :return intersecting_points:  tensor of the size of nodesDeformed where any non-negative entry indicates teh simplex the vertex is in; excludes the simplices the vertex was part of in the material configuration
            :rtype intersecting_points: torch.Tensor"""
        numberTetrahedras = torch.tensor(self.simplices.size(0), device=self.device)
        max_num = 1000
        iter = int((numberTetrahedras / max_num).floor() + 1)
        results = torch.zeros((iter, nodesDeformed.size(0)), device=self.device)
        for cnt in range(iter):
            if cnt == 0:
                simplicesTemp = self.simplices[cnt * max_num:(cnt + 1) * max_num]
                resTemp = self.intersectionDetection3D(nodesDeformed, simplicesTemp)
                resTemp = torch.where(resTemp >= 0, resTemp + (cnt * max_num), resTemp)
            elif cnt == iter - 1:
                simplicesTemp = self.simplices[cnt * max_num:]
                resTemp = self.intersectionDetection3D(nodesDeformed, simplicesTemp)
                resTemp = torch.where(resTemp >= 0, resTemp + (cnt * max_num), resTemp)
            else:
                simplicesTemp = self.simplices[cnt * max_num:(cnt + 1) * max_num]
                resTemp = self.intersectionDetection3D(nodesDeformed, simplicesTemp)
                resTemp = torch.where(resTemp >= 0, resTemp + (cnt * max_num), resTemp)
            results[cnt] = resTemp
        intersecting_points = torch.max(results, dim=0)[0]
        return intersecting_points
    def barycentricToCartesian3D(self,nodeCoordinates: Tensor, e1: Tensor, e2: Tensor, e3: Tensor, e4: Tensor)->Tensor:
        """Transformation from barycentric to cartesian coordinates in 3D.
            :param barycentricCoords: barycentric coordinates
            :type barycentricCoords: torch.Tensor
            :param e1,e2,e3,e4: cartesian coordinates of simplex vertices
            :type e1,e2,e3,e4: torch.Tensor
            :return p: cartesian coordinates
            :rtype p: torch.Tensor"""
        p = torch.unsqueeze(e1, dim=1) * nodeCoordinates[:, 0:3] + torch.unsqueeze(e2, dim=1) * nodeCoordinates[:,
                                                                                                 3:6] + torch.unsqueeze(
            e3, dim=1) * nodeCoordinates[:, 6:9] + torch.unsqueeze(e4, dim=1) * nodeCoordinates[:, 9:]
        return p
    def vertexIdToCoordinates3D(self,intersectedSimplices: Tensor, vertexCoords: Tensor)->Tensor:
        """Get cartesian coordinates of vertices in 3D.
            :param intersectedSimplices: tensor of intersected Simplices
            :type intersectedSimplices: torch.Tensor
            :param vertexCoords: tensor of vertex coordinates
            :type vertexCoords: torch.Tensor
            :return: cartesian coordinates
            :rtype: torch.Tensor"""
        intersectedSimplices = intersectedSimplices[intersectedSimplices >= 0]
        coordinates = torch.zeros((intersectedSimplices.size()[0], 12), device=self.device)
        coordinates[:, 0:3] = vertexCoords[self.simplices[intersectedSimplices.long()][:, 0].long()]
        coordinates[:, 3:6] = vertexCoords[self.simplices[intersectedSimplices.long()][:, 1].long()]
        coordinates[:, 6:9] = vertexCoords[self.simplices[intersectedSimplices.long()][:, 2].long()]
        coordinates[:, 9:] = vertexCoords[self.simplices[intersectedSimplices.long()][:, 3].long()]
        return coordinates
    def cartesianToBarycentric3D(self,nodeCoordinates: Tensor, p: Tensor)->Tuple[Tensor,Tensor,Tensor,Tensor]:
        """Transformation from cartesian to barycentric coordinates in 3D.
            :param nodeCoordinates: cartesian coordinates of simplex nodes
            :type nodeCoordinates: torch.Tensor
            :param p: cartesian coordiantes of point
            :type p: torch.Tensor
            :return e1,e2,e3,e4: barycentric coordinates of point
            :rtype: torch.Tensors"""
        a = nodeCoordinates[:, 0:3]
        b = nodeCoordinates[:, 3:6]
        c = nodeCoordinates[:, 6:9]
        d = nodeCoordinates[:, 9:]
        vbp = p - b
        vbd = d - b
        vbc = c - b
        vap = p - a
        vac = c - a
        vad = d - a
        vab = b - a
        va = (1 / 6) * torch.einsum('ij,ij->i', vbp.float(), torch.cross(vbd, vbc))
        vb = (1 / 6) * torch.einsum('ij,ij->i', vap.float(), torch.cross(vac, vad))
        vc = (1 / 6) * torch.einsum('ij,ij->i', vap.float(), torch.cross(vad, vab))
        vd = (1 / 6) * torch.einsum('ij,ij->i', vap.float(), torch.cross(vab, vac))
        v = (1 / 6) * torch.einsum('ij,ij->i', vab.float(), torch.cross(vac, vad))
        e1 = va / v
        e2 = vb / v
        e3 = vc / v
        e4 = vd / v
        return e1, e2, e3, e4
    def loss3D(self,nodesDeformed: Tensor)->Tensor:
        """Calculates the selfintersection loss in 3D.
            :param nodesDeformed: deformed mesh nodes
            :type nodesDeformed: torch.Tensor
            :return loss: summed depth of selfintersecting points
            :rtype loss: torch.Tensor"""
        intersectingNodes = self.batching3D(nodesDeformed)
        sel1=intersectingNodes>=0
        sel2=self.vertMask == 1
        sel=sel1*sel2
        if intersectingNodes[sel].size(0) == 0:
            loss = torch.zeros(1).to(device=self.device)
        else:
            loss = self.correctionForce3D(intersectingNodes, nodesDeformed)
        return loss
    def correctionForce3D(self,intersectingNodes: Tensor, nodesDeformed: Tensor)->Tensor:
        """Calculates the depth of selfintersecting points to the nearest surface in 3D in the material configuration.
            :param nodesDeformed: displaced vertices
            :type nodesDeformed: torch.Tensor
            :param intersectingSimplices: tensor of the size of nodesDeformed where any non-negative entry indicates teh simplex the vertex is in; excludes the simplices the vertex was part of in the material configuration
            :type intersectingSimplices: torch.Tensor
            :return: squared and summed depth of selfintersecting points
            :rtype: torch.Tensor"""
        sel1=intersectingNodes>=0
        sel2=self.vertMask == 1
        sel=sel1*sel2
        points = nodesDeformed[sel]
        intersectingNodes = intersectingNodes[sel]
        coordinatesDeformed = self.vertexIdToCoordinates3D(intersectingNodes,nodesDeformed)
        coordinatesOrig = self.vertexIdToCoordinates3D(intersectingNodes, self.verticesUndeformed)
        e1, e2, e3, e4 = self.cartesianToBarycentric3D(coordinatesDeformed, points)
        resamp = self.barycentricToCartesian3D(coordinatesOrig, e1, e2, e3, e4)
        inter_sdf,_,_  = self.intP (resamp, self.sdf)
        return torch.sum(torch.sqrt(inter_sdf** 2))
    def intersectionsNodes(self,nodesDeformed: Tensor)->Tensor:
        """Calculates the selfintersection loss in 3D.
            :param x: deformation parameters
            :type x: torch.tensor
            :return resulting_loss: summed depth of seflintersecting points
            :rtype resulting_loss: torch.tensor"""
        intersectingNodes = self.batching3D(nodesDeformed)
        print("here",intersectingNodes.size())
        if intersectingNodes[intersectingNodes >= 0].size(0) == 0:
            res = torch.zeros(1).to(device=self.device)
        else:
            res = self.correctionForce3D(intersectingNodes, nodesDeformed)
        selector=intersectingNodes >= 0
        return res,nodesDeformed[intersectingNodes >= 0],selector,intersectingNodes[intersectingNodes >= 0]

class selfintersectionDetection:
    """ Class for self-collision detection based on meshes.
        :param pts: mesh points used for collision detection
        :type pts: torch.Tensor
        :param sdf: signed distance field of mesh in material configuration
        :type sdf: torch.Tensor
        :param interpolator: interpolator for pts locations after displacement in SDF
        :type interpolator: nearest, linear, or cubic interpolation class
        :param simplices: simplices of mesh (either triangles or tetrahedra); (#of simplices,3 or 4)
        :type simplices: torch.Tensor
        :param verticesUndeformed: vertices of mesh (# of vertices, 2 or 3) in material configuration
        :type verticesUndeformed: torch.Tensor
        :param coef: coefficient applied to the collision loss
        :type coef: float
        :param device: computation device, see torch docs; either cpu or cuda
        :type device: string
        :param pointsMask: a mask for pts if only a subset of pts needs to be checked for collision
        :type pointsMask: torch.Tensor
        :param pointsMaskLabel: the mask label for pointsMask that needs to be checked for collision
        :type pointsMaskLabel: int
        :param unmaskedDis: if the object is modelled as two moving pieces this displacement is used in the partners collision detection
        :type unmaskedDis: torch.Tensor
        :param vertMask: a mask for vertices if only a subset of vertices needs to be checked for collision
        :type vertMask: torch.Tensor
        :param vertMaskLabel: the mask label for vertMask that needs to be checked for collision
        :type vertMaskLabel: int
        :param siPartner: a collision partner if the object is modelled as two moving pieces
        :type siPartner: selfintersectionDetection2D or selfintersectionDetection3D class
        :return: self intersection class depending on the dimensionality
        :rtype: selfintersectionDetection2D or selfintersectionDetection3D class
        """
    def __new__(cls, pts: Tensor,sdf: Tensor,interpolator: Union[nearest,linear,cubic],simplices: Tensor,verticesUndeformed: Tensor,coef: float,device: str,pointsMask: Optional[Tensor] = None,pointsMaskLabel: Optional[int] = None,unmaskedDis: Optional[Tensor] = None,vertMask: Optional[Tensor] = None,vertMaskLabel: Optional[int] = None,siPartner: Optional[Union[selfintersectionDetection2D,selfintersectionDetection3D]] = None)->Union[selfintersectionDetection2D,selfintersectionDetection3D]:
        """ Constructor method. Decides dimensionality."""
        if pts.size(1) == 2:
            return selfintersectionDetection2D(pts,sdf,interpolator,simplices,verticesUndeformed,coef,device,pointsMask,pointsMaskLabel,unmaskedDis,vertMask,vertMaskLabel,siPartner)
        elif pts.size(1) == 3:
            return selfintersectionDetection3D(pts,sdf,interpolator,simplices,verticesUndeformed,coef,device,pointsMask,pointsMaskLabel,unmaskedDis,vertMask,vertMaskLabel,siPartner)
        else:
            raise ValueError("Unsupported dimension. Only 2D and 3D are supported.")


import torch
from typing import Optional, Type, Union
from torch import Tensor
from ..interpolation import *

class intersectionDetection:
    """ Class for collision detection based on signed distance fields (SDF). The deformable object is represented as a point cloud, the non-deformable obstacle as an SDF.
    :param pts: points representing the deformable object
    :type pts: torch.Tensor
    :param sdf: SDF of non-deformable obstacles
    :type sdf: torch.Tensor
    :param interpolator: interpolator for pts locations after displacement in SDF
    :type interpolator: nearest, linear, or cubic interpolation class
    :param coef: coefficient applied to the collision loss
    :type coef: float
    :param pointsMask: a mask for pts if only a subset of pts needs to be checked for collision
    :type pointsMask: torch.Tensor
    :param pointsMaskLabel: the mask label for pointsMask that needs to be checked for collision
    :type pointsMaskLabel: int
    """
    def __init__(self, pts: Tensor,sdf: Tensor,interpolator: Union[nearest,linear,cubic], coef: float=1.,pointsMask: Optional[Tensor] = None,pointsMaskLabel: Optional[int] = None):
        """Constructor method."""
        self.sdf = sdf
        self.interpolator=interpolator
        self.coef=coef
        self.pts=pts
        self.pointsMask=pointsMask
        self.pointsMaskLabel=pointsMaskLabel
    def __call__(self,dis: Tensor=None,**kwargs)->Tensor:
        """ Calculates the summed depth of intersecting points. Adds tiny to prevent exploding gradients.
        If pointsMask is given, only masked pts are checked.
        :param dis: displacement of pts
        :type dis: torch.Tensor
        :return: summed depth of intersecting points
        :rtype: torch.Tensor
        """
        if self.pointsMask is not None:
            dis=dis[self.pointsMask==self.pointsMaskLabel]
            pts=self.pts[self.pointsMask==self.pointsMaskLabel]
            sdf_int, _, _ = self.interpolator(pts + dis, self.sdf)
        else:
            sdf_int,_,_=self.interpolator(self.pts+dis, self.sdf)
        loss = torch.sqrt((sdf_int** 2)+ torch.finfo().tiny)
        return torch.sum(loss)*self.coef





import torch
import tqdm
from typing import Optional, Type, Union, Tuple
from torch import Tensor
from ..geometricTransformations import *
from .optim_gradientDescentBacktracking import *

class closureGradientDescent:
    """ Class for a custom closure function (see pytorch docs) to evaluate the loss function.
        :param optimizer: optimizer class used for the minimization problem
        :type optimizer: gradientDescentBacktracking or torch.optim.Optimizer class'
        :param transformer: transformer
        :type transformer: affineTransformation or nonrigidDeformation class
        :param mainTerms: main terms of the cost function, for example similarity measure or collision detection
        :type mainTerms: list of any regularization, collisionDetection, or similarityMeasure class
        :param regTerms: regularization terms of the cost function
        :type regTerms: list of any regularization, collisionDetection, or similarityMeasure class
            """
    def __init__(self,optimizer: Union[gradientDescentBacktracking,torch.optim.Optimizer],transformer: Union[affineTransformation,nonrigidDeformation],mainTerms: Optional[list]=None,regTerms: Optional[list]=[]):
        """Constructor method.                """
        self.optimizer=optimizer
        self.mainTerms=mainTerms
        self.regTerms=regTerms
        self.transformer=transformer

    def __call__(self,decisionVariables: Tensor)->Tensor:
        """ Initiates the forward pass. In case of gradients switched on, also zeros the gradients first.
                :param decisionVariables: Tensor with decision variables
                :type decisionVariables: torch.Tensor
                :return loss: backpropagated accumulated loss
                :rtype loss: torch.Tensor
                """
        if decisionVariables.requires_grad:
            self.optimizer.zero_grad()
            _,loss=self.transformer(decisionVariables,self.mainTerms,self.regTerms)
            loss.backward()
        else:
            _, loss = self.transformer(decisionVariables, self.mainTerms, self.regTerms)
        return loss

def algorithmGradientDescent(iterations: int,optimizerF: Union[gradientDescentBacktracking,torch.optim.Optimizer],closureF: Type,decisionVariables: Tensor)->dict:
    """ Starts a default iterative, gradient descent optimization scheme.
            :param iterations: number of steps in scheme
            :type iterations: int
            :param optimizerF: optimizer class used for the minimization problem
            :type optimizerF: gradientDescentBacktracking or torch.optim.Optimizer class
            :param closureF: closure function to calculate loss and backpropagate.
            :type closureF: either a function or a class with a __call__ method
            :param decisionVariables: Tensor with decision variables
            :type decisionVariables: torch.Tensor
            :return dict: dictionary with gradient history and objective function history
            :rtype dict: dict
            """
    gradientFHistory=[]
    objectiveLossFHistory=[]
    for i in tqdm.tqdm(range(iterations),desc="Progress"):
        loss=optimizerF.step(lambda: closureF(decisionVariables))
        gradientFHistory.append(torch.norm(decisionVariables.grad).item())
        objectiveLossFHistory.append(loss.detach().cpu().item())
        dict = {"gradientHistory": gradientFHistory,
            "objectiveLossHistory": objectiveLossFHistory}
    return dict


def algorithmGradientDescentStochastic(device: str,iterations: int,optimizerF: Union[gradientDescentBacktracking,torch.optim.Optimizer],closureF: Type,decisionVariables: Tensor,evalPoints: Tensor,evalPointsIntensities: Tensor,percentage: float,stochasticTerms: list)->dict:
    """ Starts a default iterative, gradient descent optimization scheme. It uses stochastic selection of evaluation points.
            :param device: computation device, see torch docs
            :type device: str
            :param iterations: number of steps in scheme
            :type iterations: int
            :param optimizerF: optimizer class used for the minimization problem
            :type optimizerF: gradientDescentBacktracking or torch.optim.Optimizer class
            :param closureF: closure function to calculate loss and backpropagate.
            :type closureF: either a function or a class with a __call__ method
            :param decisionVariables: Tensor with decision variables
            :type decisionVariables: torch.Tensor
            :param evalPoints: range of evaluation points available for selection
            :type evalPoints: torch.Tensor
            :param evalPointsIntensities: range of intensities of evaluation points available for selection
            :type evalPointsIntensities: torch.Tensor
            :param percentage: percentage of points used in each iteration ranging from [0;1]
            :type percentage: float
            :param stochasticTerms: list of mainTerms and regTerms that need updates to their points in every iteration
            :type stochasticTerms: list
            :return dict: dictionary with gradient history and objective function history
            :rtype dict: dict
            """
    gradientFHistory=[]
    objectiveLossFHistory=[]
    length=evalPoints.size(0)
    numberOfPoints=int(length*percentage)
    for i in tqdm.tqdm(range(iterations),desc="Progress"):
        random_tensor = torch.randperm(evalPoints.size(0))[:numberOfPoints]

        evalPointsStochastic = evalPoints[random_tensor].to(device=device)
        evalPointsIntensitiesStochastic = evalPointsIntensities[random_tensor].to(device=device)

        for term in stochasticTerms:
            if hasattr(term, 'pts'):
                term.pts=evalPointsStochastic
            if hasattr(term, 'intensities'):
                term.intensities=evalPointsIntensitiesStochastic

        loss=optimizerF.step(lambda: closureF(decisionVariables))
        gradientFHistory.append(torch.norm(decisionVariables.grad).item())
        objectiveLossFHistory.append(loss.detach().cpu().item())
        dict = {"gradientHistory": gradientFHistory,
            "objectiveLossHistory": objectiveLossFHistory}
    return dict



import matplotlib.pyplot as plt
import torch
from scipy import ndimage
import math
import numpy as np
import skfmm
from ..interpolation import *
from typing import Optional, Type, Union, Tuple
from torch import Tensor

def getEvaluationPoints(device: str,image: Tensor,mask: Optional[Tensor]=None,maskLabel: Optional[Tensor]=None,dilation: Optional[int]=None,exteriorLayers: Optional[int]=None,percentageOfPoints: Optional[float]=None,random: Optional[bool]=None)->Tensor:
    """Returns a set of evaluation points representing the domains to be registered.
    :param device: computation device, see torch docs
    :type device: str
    :param image: moving image of the registration scene
    :type image: torch.Tensor
    :param mask: if only a certain subdomain is registered this mask of moving image is needed
    :type mask: torch.Tensor
    :param maskLabel: if only a certain subdomain is registered the according label is needed
    :type maskLabel: torch.Tensor
    :param dilation: dilating the mask for int steps
    :type dilation: int
    :param exteriorLayers: only this number of exterior layer of the subdomain are considered
    :type exteriorLayers: int
    :param percentageOfPoints: percentage of points selected from the domain (from 0 to 1)
    :type percentageOfPoints: float
    :param random: switch to randomly select the percentage of points
    :type random: bool
    """
    indices = np.indices(image.cpu()[0].size())
    pts = np.empty((np.prod(image.cpu().size()), len(image[0].cpu().size())))
    for i, slide in enumerate(indices):
        pts[:, i] = slide.flatten()
    pts = torch.from_numpy(pts).to(device=device).float()
    if mask is not None:
        if dilation is not None:
            mask = torch.where(mask == maskLabel, 1, 0)
            mask=torch.from_numpy(ndimage.binary_dilation(mask.cpu(),iterations=dilation)).to(device=device)
            mask=torch.where(mask==1,maskLabel,maskLabel-1).to(device=device)
        if exteriorLayers is not None:
            mask = torch.where(mask == maskLabel, 1, 0).to(device=device)
            exteriorLayer = torch.unsqueeze(torch.from_numpy(
                ndimage.binary_erosion(mask[0].cpu().numpy().astype(np.bool_), iterations=exteriorLayers)).to(
                device=device).int(), dim=0)
            mask = mask - exteriorLayer
            mask=torch.where(mask==1,maskLabel,maskLabel-1)
        pts = pts[mask.flatten() == maskLabel]
    if percentageOfPoints is not None:
        numberOfPointsInt = int(pts.size(0) * percentageOfPoints)
        if random:
            random_tensor = torch.randperm(pts.size(0))[:numberOfPointsInt]
            pts = pts[random_tensor]
        else:
            skip = getSkipInterval(pts, numberOfPointsInt)
            pts = pts[::skip]
    return pts

def getSkipInterval(pts: Tensor,number: int)->int:
    """Calculates the necessary interval between evaluation points to achieve the required amount.
        :param pts: all points
        :type pts: torch.Tensor
        :param number: approximate number of evaluation points required
        :type number: int
        :return skip: interval between points
        :rtype skip: int"""
    ptsNumber=pts.size(0)
    skip=math.ceil(ptsNumber/number)
    if skip==0:
        return 1
    else:
        return skip

def getGridPoints(movingImage: Tensor,scale: Tensor,timesteps: int=1)->Tensor:
    """Returns the control point grid for velocity field interpolation.
        :param movingImage: moving image (1,dim1,dim2 (,dim3))
        :type movingImage: torch.Tensor
        :param scale: tensor with stepsize between two consecutive control points in each dimension in pixel
        :type scale: torch.Tensor
        :param timesteps: time steps into which the interval t=[0;1] is divided
        :type timesteps: int
        :return skip: control point grid
        :rtype skip: torch.Tensor"""
    device=movingImage.device
    if scale.size(0)==2:
        x = torch.zeros((timesteps, 2, int(movingImage.size(1) / scale[0]) + 1, int(movingImage.size(2) / scale[1]) + 1))
        return x.to(device=device)
    elif scale.size(0)==3:
        x = torch.zeros((timesteps, 3, int(movingImage.size(1) / scale[0]) + 1, int(movingImage.size(2) / scale[1]) + 1, int(movingImage.size(3) / scale[2]) + 1))
        return x.to(device=device)
    else:
        raise Exception("wrong dimension: scale & moving image")


def assignPoints(device: str,pts: Tensor,mask: Tensor,segments: Tensor,initialValue: Optional[float]=10000)->Tensor:
    """Given a mask, points are assigned to one of the entries in segments. This process either checks whether a point lies inside a subdomain or to which it is closest to.
        :param device: computation device, see torch docs
        :type device: str
            :param pts: points
        :type pts: torch.Tensor
        :param mask: mask (1,dim1,dim2 (,dim3))
        :type mask: torch.Tensor
        :param segments: segments to which points are assigned to
        :type segments: torch.Tensor
        :param initialValue: initial distance value
        :type initialValue: float
        :return s: segmentation
        :rtype s: torch.Tensor"""
    segmentMasks=[]
    for s in segments:
        segmentMasks.append(torch.where(mask == s, 1, 0))
    interTemp = nearest(device, torch.tensor([1., 1., 1.], device=device))
    s = interTemp(pts, mask)
    s = s.flatten()
    set_Elements = set(torch.unique(s).tolist())
    unique_elements = set_Elements.symmetric_difference(segments)
    if len(unique_elements) != 0:
        wrongVals = unique_elements
        inter = linear(device, torch.tensor([1., 1., 1.], device=device))
        sdfs=[]
        for m in segmentMasks:
            m = torch.where(m == 1, -1, 1)
            sdfs.append(torch.unsqueeze(torch.from_numpy(skfmm.distance(m.cpu()[0])).to(device=device), dim=0))
        for val in wrongVals:
            v = pts[s == val]
            decisionMatrix=torch.ones(v.size(0),device=device)*initialValue
            for i,sdf in enumerate(sdfs):
                v1, _, _ = inter(v, sdf)
                decisionMatrix=torch.where(v1.flatten()<decisionMatrix,segments[i],decisionMatrix)
            s[s == val] = decisionMatrix
    return s




from skimage import measure
from .. import  geometricTransformations,  collisionDetection,utils, interpolation
import matplotlib.pyplot as plt
import igl
from .postprocessing_utils import *
from shapely.geometry import LineString, Polygon
from shapely.ops import polygonize, unary_union
from typing import Optional, Type, Union, Tuple
from torch import Tensor
import skimage
def measure_selfIntersection3d(device: str,workingDirectory: str,voxelToMm: Optional[Tensor]=None,segmentsOfInterest: Optional[Tensor]=None,vertices: Tensor=None,simplices: Tensor=None)->Tensor:
    """ Calculates the overlap from selfintersections in 3D. Object is input as vertex and simplex collection.
        :param device: sets the computation device, see torch
        :type device: string
        :param workingDirectory: path to working directory, see docs
        :type workingDirectory: string
        :param segmentsOfInterest: segmentations of interest list
        :type segmentsOfInterest: list
        :param voxelToMm: cell dimensions in mm
        :type voxelToMm: torch.Tensor
        :param vertices: vertex collection
        :type vertices: list
        :param simplices: simplex collection
        :type simplices: list
        :return : overlap
        :rtype : Tensor
    """
    # BASICS: load images
    movingImageMask = torch.unsqueeze(torch.from_numpy(np.load(os.path.join(workingDirectory, "moving_mask.npy"))),
                                      dim=0).to(device=device)

    indices = np.indices(movingImageMask.cpu()[0].size())
    coords = np.empty((np.prod(movingImageMask.cpu().size()), len(movingImageMask[0].cpu().size())))
    for i, slide in enumerate(indices):
        coords[:, i] = slide.flatten()
    coords = torch.from_numpy(coords).to(device=device).float()

    verticesSegmentation=utils.assignPoints(device, vertices, movingImageMask, segmentsOfInterest, initialValue=10000).long()

    vertices = checkAffine(device, workingDirectory, vertices, verticesSegmentation)
    vertices = checkNonrigid(device, workingDirectory, vertices, verticesSegmentation)

    verticesNumpy = vertices.cpu().numpy().copy(order='C')
    wn = igl.fast_winding_number_for_meshes(verticesNumpy, simplices.cpu().numpy(), coords.cpu().numpy())
    wn = np.round(wn, decimals=1)

    coordSel = coords.cpu().numpy()[wn > 1]

    if voxelToMm is not None:
        volume = coordSel.shape[0] * np.prod(voxelToMm.cpu().numpy())
    else:
        volume = coordSel.shape[0]
    return volume


def measure_selfIntersection2d(device: str,workingDirectory: str,voxelToMm: Optional[Tensor]=None,segmentsOfInterest: Optional[list]=None)->Tensor:
    """ Calculates the overlap from selfintersections in 2D. Object is input as contour.
        :param device: sets the computation device, see torch
        :type device: string
        :param workingDirectory: path to working directory, see docs
        :type workingDirectory: string
        :param segmentsOfInterest: segmentations of interest list
        :type segmentsOfInterest: list
        :param voxelToMm: cell dimensions in mm
        :type voxelToMm: torch.Tensor
        :return : overlap
        :rtype : Tensor
    """
    movingImageMask = torch.unsqueeze(torch.from_numpy(np.load(os.path.join(workingDirectory, "moving_mask.npy"))),
                                      dim=0).to(device=device)
    inter = interpolation.nearest(device, scale=torch.tensor([1., 1.], device=device))
    movingImageMaskCont = torch.from_numpy(np.load(os.path.join(workingDirectory, "moving_mask.npy"))).to(device=device)
    mask=torch.zeros(movingImageMaskCont.size(),device=device)

    for segment in segmentsOfInterest:
        maskTemp=torch.where(movingImageMaskCont==segment,1,0).to(device=device)
        mask=mask+maskTemp
    print(mask.size())

    contour=skimage.measure.find_contours(mask.cpu().numpy(),level=0.5)[0]
    contour=torch.from_numpy(contour).to(device=device)


    segmentation = inter(contour, movingImageMask)

    contour = checkAffine(device, workingDirectory, contour, segmentation)
    contour = checkNonrigid(device, workingDirectory, contour, segmentation)

    line1 = LineString(contour.cpu().numpy())
    if line1.is_closed:
        polygon1 = Polygon(line1)
    else:
        raise Exception("Postprocessing contour not closed")
    lines = LineString(polygon1.exterior.coords)
    intersections = unary_union(
        [lines.intersection(LineString([lines.coords[i], lines.coords[i + 1]])) for i in range(len(lines.coords) - 1)])
    intersecting_areas = list(polygonize(intersections))
    sum = 0
    i = 0
    listIntersection=[]
    for poly in intersecting_areas:
        listIntersection.append(poly.area)
        sum = sum + poly.area
        i = i + 1
        print(i, poly.area)

        fig, ax = plt.subplots()
        ax.imshow(movingImageMask.cpu()[0] * 0, cmap="binary")
        ax.plot(contour[:, 1].cpu(), contour[:, 0].cpu(), c='r', )
        ax.set_xticks([])  # Remove x-axis ticks
        ax.set_yticks([])  # Remove y-axis ticks
        x, y = poly.exterior.xy
        ax.fill(y, x, facecolor="green", edgecolor="green", alpha=1)
        ax.set_xlim([10, 90])  # Limit x-axis to range [30, 70]
        ax.set_ylim([20, 80])
        plt.savefig(workingDirectory + f'/results/synthetic_fixed_SI_{i}_ccdir.eps', format='eps',
                    bbox_inches='tight')
        plt.close()

    print(f"The area of the self-intersecting region is: {sum}")
    return listIntersection









import torch
from typing import Optional, Type, Union, Tuple
from torch import Tensor
from ..interpolation import *

class forwardEuler:
    """ Class for forward euler numerical integration of velocity fields. Time interval is assumed to be [0;1] and step size is assumed to be equidistant.
    :param t_steps: steps taken to cover time interval
    :type t_steps: int
    :param stationary: flag for stationary velocity fields
    :type stationary: bool
    """
    def __init__(self,t_steps: int,stationary: Optional[bool] = False):
        """ Constructor method.        """
        self.t_steps=t_steps
        self.stationary=stationary

    def __call__(self,pts: Tensor,velocityField: Tensor,interpolator: Union[nearest,linear,cubic],mainTerms: list,regTerms: list)->Tuple[Tensor,Tensor]:
        """ Numerical integration. Decides whether to treat the velocity field as stationary or non-stationary.
        :param pts: points coordinates in material configuration
        :type pts: torch.Tensor
        :param velocityField: velocity field (time frames, dimensions, # control points dim 1, # control points dim 2 [,# control points dim 3])
        :type velocityField: torch.Tensor
        :param interpolator:  interpolation method for velocity fields
        :type interpolator: nearest, linear, or cubic interpolation class
        :param mainTerms: main terms of the cost function, for example similarity measure or collision detection
        :type mainTerms: list of any regularization, collisionDetection, or similarityMeasure class
        :param regTerms: regularization terms of the cost function
        :type regTerms: list of any regularization, collisionDetection, or similarityMeasure class
        :return: Tuple of displaced points and loss
        :rtype: (torch.Tensor,torch.Tensor)
        """
        if self.stationary:
            return self.callStationary(pts,velocityField,interpolator,mainTerms,regTerms)
        else:
            return self.callNonstationary(pts,velocityField,interpolator,mainTerms,regTerms)

    def callNonstationary(self,pts: Tensor,velocityField: Tensor,interpolator: Union[nearest,linear,cubic],mainTerms: list,regTerms: list)->Tuple[Tensor,Tensor]:
        """ Nonstationary numerical integration. See __call__.
        """
        if mainTerms is not None or regTerms is not None:
            loss=0
            pts_orig=pts.clone()
            for i in range(self.t_steps):
                vel,vel_jac,vel_lap=interpolator(pts,velocityField[i])
                dis=vel.t()*(1/self.t_steps)
                dis_jac=vel_jac*(1/self.t_steps)
                for term in regTerms:
                    loss=loss+term(vel=vel.t(),vel_jac=vel_jac,vel_lap=vel_lap,dis=dis,dis_jac=dis_jac)*(1/self.t_steps)
                pts = pts + dis
                #loss=loss+mainTerm(vel=vel.t(),vel_jac=vel_jac,vel_lap=vel_lap,dis=dis,dis_jac=dis_jac) #if data fidelty term is inside time integral in continuous formulation of the cost function
            for main in mainTerms:
                loss=loss+main(dis=pts-pts_orig)
            return pts,loss
        else:
            for i in range(self.t_steps):
                vel,jac,lap=interpolator(pts,velocityField[i])
                pts = pts + vel.t() * (1 / self.t_steps)
            return pts,None
    def callStationary(self,pts: Tensor,velocityField: Tensor,interpolator: Union[nearest,linear,cubic],mainTerms: list,regTerms: list)->Tuple[Tensor,Tensor]:
        """ Stationary numerical integration. See __call__.
        """
        if mainTerms is not None or regTerms is not None:
            loss=0
            pts_orig=pts.clone()
            for i in range(self.t_steps):
                vel,vel_jac,vel_lap=interpolator(pts,velocityField[0])
                dis=vel.t()*(1/self.t_steps)
                dis_jac=vel_jac*(1/self.t_steps)
                for term in regTerms:
                    loss=loss+term(vel=vel.t(),vel_jac=vel_jac,vel_lap=vel_lap,dis=dis,dis_jac=dis_jac)*(1/self.t_steps)
                pts = pts + dis
            for main in mainTerms:
                loss=loss+main(dis=pts-pts_orig)
            return pts,loss
        else:
            for i in range(self.t_steps):
                vel,jac,lap=interpolator(pts,velocityField[0])
                pts = pts + vel.t() * (1 / self.t_steps)
            return pts,None

class trapezoidal:
    """ Class for forward euler trapezoidal integration of velocity fields. Time interval is assumed to be [0;1] and step size is assumed to be equidistant.
    :param t_steps: steps taken to cover time interval
    :type t_steps: int
    :param stationary: flag for stationary velocity fields
    :type stationary: bool
    :param corrector_steps: maximum iterations for the predictor-corrector method
    :type corrector_steps: int
    :param tol: tolerance for predictor-corrector method
    :type tol: float
    """
    def __init__(self,t_steps: int,stationary=False,corrector_steps: int=1,tol: float=0.0001):
        """ Constructor method.
        """
        self.t_steps=t_steps
        self.stationary=stationary
        self.corrector_steps=corrector_steps
        self.tol=tol

    def __call__(self,pts: Tensor,velocityField: Tensor,interpolator: Union[nearest,linear,cubic],mainTerms: list,regTerms: list)->Tuple[Tensor,Tensor]:
        """ Numerical integration. Decides whether to treat the velocity field as stationary or non-stationary.
        :param pts: points coordinates in material configuration
        :type pts: torch.Tensor
        :param velocityField: velocity field (time frames, dimensions, # control points dim 1, # control points dim 2 [,# control points dim 3])
        :type velocityField: torch.Tensor
        :param interpolator:  interpolation method for velocity fields
        :type interpolator: nearest, linear, or cubic interpolation class
        :param mainTerms: main terms of the cost function, for example similarity measure or collision detection
        :type mainTerms: list of any regularization, collisionDetection, or similarityMeasure class
        :param regTerms: regularization terms of the cost function
        :type regTerms: list of any regularization, collisionDetection, or similarityMeasure class
        :return: Tuple of displaced points and loss
        :rtype: (torch.Tensor,torch.Tensor)
        """
        if self.stationary:
            return self.callStationary(pts,velocityField,interpolator,mainTerms,regTerms)
        else:
            return self.callNonstationary(pts,velocityField,interpolator,mainTerms,regTerms)

    def callNonstationary(self,pts: Tensor,velocityField: Tensor,interpolator: Union[nearest,linear,cubic],mainTerms: list,regTerms: list)->Tuple[Tensor,Tensor]:
        """ Nonstationary numerical integration. See __call__.
        """
        if mainTerms is not None or regTerms is not None:
            loss=0
            pts_orig=pts.clone()
            for i in range(self.t_steps - 1):
                vel, vel_jac, vel_lap = interpolator(pts, velocityField[i])
                dis = vel.t() * (1 / (self.t_steps - 1))
                dis_jac=vel_jac*(1 / (self.t_steps - 1))
                pts_for = pts + dis
                pts_for_old = pts_for

                for j in range(self.corrector_steps):
                    vel_i_1, vel_jac_i_1, vel_lap_i_1 = interpolator(pts_for, velocityField[i + 1])
                    dis_i_1 = (vel.t() + vel_i_1.t()) * (1 / (self.t_steps - 1)) * 0.5
                    dis_jac_i_1 = (vel_jac+ vel_jac_i_1) * (1 / (self.t_steps - 1)) * 0.5
                    pts_for = pts + dis_i_1
                    for term in regTerms:
                        loss = loss + (term(vel=vel.t(),vel_jac=vel_jac,vel_lap=vel_lap,dis=dis,dis_jac=dis_jac)+term(vel=vel_i_1.t(), vel_jac=vel_jac_i_1, vel_lap=vel_lap_i_1, dis=dis_i_1, dis_jac=dis_jac_i_1))*0.5 * (1 / self.t_steps)
                    if torch.norm(pts_for.data - pts_for_old.data) < self.tol:
                        break
                    else:
                        pts_for_old = pts_for
                pts = pts_for
            for main in mainTerms:
                loss=loss+main(dis=pts-pts_orig)
            return pts,loss
        else:
            for i in range(self.t_steps-1):
                vel,vel_jac,vel_lap=interpolator(pts,velocityField[i])
                dis=vel.t()*(1 / (self.t_steps - 1))
                pts_for = pts + dis
                pts_for_old=pts_for
                for j in range(self.corrector_steps):
                    vel_i_1, vel_jac_i_1, vel_lap_i_1 = interpolator(pts_for, velocityField[i + 1])
                    dis_i_1 = (vel.t() + vel_i_1.t()) * (1 / (self.t_steps - 1)) * 0.5
                    pts_for = pts + dis_i_1
                    if torch.norm(pts_for.data - pts_for_old.data) < self.tol:
                        break
                    else:
                        pts_for_old = pts_for
                pts = pts_for
            return pts,None
    def callStationary(self,pts: Tensor,velocityField: Tensor,interpolator: Union[nearest,linear,cubic],mainTerms: list,regTerms: list)->Tuple[Tensor,Tensor]:
        """ Stationary numerical integration. See __call__.
        """
        if mainTerms is not None or regTerms is not None:
            loss = 0
            pts_orig=pts.clone()
            for i in range(self.t_steps - 1):
                vel, vel_jac, vel_lap = interpolator(pts, velocityField[0])
                dis = vel.t() * (1 / (self.t_steps - 1))
                dis_jac = vel_jac * (1 / (self.t_steps - 1))
                pts_for = pts + dis
                pts_for_old = pts_for
                for j in range(self.corrector_steps):
                    vel_i_1, vel_jac_i_1, vel_lap_i_1 = interpolator(pts_for, velocityField[0])
                    dis_i_1 = (vel.t() + vel_i_1.t()) * (1 / (self.t_steps - 1)) * 0.5
                    dis_jac_i_1 = (vel_jac + vel_jac_i_1) * (1 / (self.t_steps - 1)) * 0.5
                    pts_for = pts + dis_i_1
                    for term in regTerms:
                        loss = loss + (term(vel=vel.t(),vel_jac=vel_jac,vel_lap=vel_lap,dis=dis,dis_jac=dis_jac) + term(vel=vel_i_1.t(), vel_jac=vel_jac_i_1, vel_lap=vel_lap_i_1, dis=dis_i_1, dis_jac=dis_jac_i_1)) * 0.5 * (
                                           1 / self.t_steps)
                    if torch.norm(pts_for.data - pts_for_old.data) < self.tol:
                        break
                    else:
                        pts_for_old = pts_for
                pts = pts_for
            for main in mainTerms:
                loss=loss+main(dis=pts-pts_orig)
            return pts, loss
        else:
            for i in range(self.t_steps - 1):
                vel, vel_jac, vel_lap = interpolator(pts, velocityField[0])
                dis = vel.t() * (1 / self.t_steps)
                pts_for = pts + dis
                pts_for_old = pts_for
                for j in range(self.corrector_steps):
                    vel_i_1, vel_jac_i_1, vel_lap_i_1 = interpolator(pts_for, velocityField[0])
                    dis_i_1 = (vel.t() + vel_i_1.t()) * (1 / (self.t_steps - 1)) * 0.5
                    pts_for = pts + dis_i_1
                    if torch.norm(pts_for.data - pts_for_old.data) < self.tol:
                        break
                    else:
                        pts_for_old = pts_for
                pts = pts_for
            return pts, None





import skimage.measure
import torch
import numpy as np
import os
from .. import interpolation, geometricTransformations,  numericalIntegration
import re
import matplotlib.pyplot as plt
import pickle
from .postprocessing_utils import *
from typing import Optional, Type, Union, Tuple
from torch import Tensor
def visual_vector(device: str,workingDirectory: str,voxelToMm: Optional[list]=None,segmentsOfInterest: Optional[list]=None,dimension: int=None,slice: int=None):
    """ Plots the displacement vector image.
        :param device: sets the computation device, see torch
        :type device: string
        :param workingDirectory: path to working directory, see docs
        :type workingDirectory: string
        :param segmentsOfInterest: segmentations of interest list
        :type segmentsOfInterest: list
        :param voxelSizes: cell dimensions in mm
        :type voxelSizes: torch.Tensor
        :param dimension: if 3D registration, set with dimension should be displayed
        :type dimension: int
        :param slice: if 3D registration, set with slice in dimension should be displayed
        :type slice: int
    """

    #BASICS: load images
    movingImageMask=torch.unsqueeze(torch.from_numpy(np.load(os.path.join(workingDirectory, "moving_mask.npy"))), dim=0).to(device=device)
    movingImage=torch.unsqueeze(torch.from_numpy(np.load(os.path.join(workingDirectory, "moving.npy"))), dim=0).to(device=device)

    indices = np.indices(movingImage.cpu()[0].size())
    pts = np.empty((np.prod(movingImage.cpu().size()), len(movingImage[0].cpu().size())))
    for i, slide in enumerate(indices):
        pts[:, i] = slide.flatten()
    pts=torch.from_numpy(pts).to(device=device).float()
    pts_orig=pts.clone()

    if segmentsOfInterest is not None:
        inter = interpolation.nearest(device,
                                      scale=torch.ones(pts.size(1), device=device))
        ptsSegmentation = inter(pts, movingImageMask).flatten().long()

        mask = torch.zeros_like(ptsSegmentation, dtype=torch.bool)
        for segment in segmentsOfInterest:
            mask |= (ptsSegmentation == segment)
        pts=pts[mask]
        ptsSegmentation=ptsSegmentation[mask]
        pts=checkAffine(device,workingDirectory,pts,ptsSegmentation)
        pts=checkNonrigid(device,workingDirectory,pts,ptsSegmentation)
        pts_orig=pts_orig[mask]
    else:
        pts = checkAffine(device, workingDirectory, pts)
        pts = checkNonrigid(device, workingDirectory, pts)

    if dimension is not None and slice is not None:
        fig, ax = plt.subplots(1)
        if dimension==0:
            pts = pts[pts_orig[:,0] == slice]
            pts_orig = pts_orig[pts_orig[:,0] == slice]
            movingImage =movingImage[0,slice,:,:]
            pts1 = pts[:, 1]
            pts2 = pts[:, 2]
            pts = torch.stack((pts1, pts2), dim=1)
            pts_orig1 = pts_orig[:, 1]
            pts_orig2 = pts_orig[:, 2]
            pts_orig = torch.stack((pts_orig1, pts_orig2), dim=1)
            move=pts-pts_orig
            extent = [0, movingImage.size(1) * voxelToMm[2].cpu(), 0, movingImage.size(0)  * voxelToMm[1].cpu()]
            pts_orig[:,0]=pts_orig[:,0]*voxelToMm[1]
            pts_orig[:,1]=pts_orig[:,1]*voxelToMm[2]
            move[:,0]=move[:,0]*voxelToMm[1]
            move[:,1]=move[:,1]*voxelToMm[2]
        elif dimension==1:
            pts = pts[pts_orig[:,1] == slice]
            pts_orig = pts_orig[pts_orig[:,1] == slice]
            movingImage = movingImage[0, :, slice, :]
            pts1 = pts[:, 0]
            pts2 = pts[:, 2]
            pts = torch.stack((pts1, pts2), dim=1)
            pts_orig1 = pts_orig[:, 0]
            pts_orig2 = pts_orig[:, 2]
            pts_orig = torch.stack((pts_orig1, pts_orig2), dim=1)
            move=pts-pts_orig
            extent = [0, movingImage.size(1) * voxelToMm[2].cpu(), 0, movingImage.size(0)  * voxelToMm[0].cpu()]
            pts_orig[:,0]=pts_orig[:,0]*voxelToMm[0]
            pts_orig[:,1]=pts_orig[:,1]*voxelToMm[2]
            move[:,0]=move[:,0]*voxelToMm[0]
            move[:,1]=move[:,1]*voxelToMm[2]

        elif dimension==2:
            pts = pts[pts_orig[:,2] == slice]
            pts_orig = pts_orig[pts_orig[:,2] == slice]
            movingImage = movingImage[0, :, :, slice]
            pts1 = pts[:, 0]
            pts2 = pts[:, 1]
            pts = torch.stack((pts1, pts2), dim=1)
            pts_orig1 = pts_orig[:, 0]
            pts_orig2 = pts_orig[:, 1]
            pts_orig = torch.stack((pts_orig1, pts_orig2), dim=1)
            move=pts-pts_orig
            extent = [0, movingImage.size(1) * voxelToMm[1].cpu(), 0, movingImage.size(0)  * voxelToMm[0].cpu()]
            pts_orig[:,0]=pts_orig[:,0]*voxelToMm[0]
            pts_orig[:,1]=pts_orig[:,1]*voxelToMm[1]
            move[:,0]=move[:,0]*voxelToMm[0]
            move[:,1]=move[:,1]*voxelToMm[1]

        else:
            raise Exception("Postprocessing check dimensions")

        ax.imshow(movingImage.cpu(),extent=extent,origin='lower')
        ax.set_title(f"vector; dim {dimension} slice {slice}")
        ax.quiver(pts_orig[:, 1].cpu(), pts_orig[:, 0].cpu(), move[:, 1].cpu(), move[:, 0].cpu(),color='r', angles='xy', scale_units='xy', scale=1)
        plt.show()
    else:
        extent = [0, movingImage.size(2) * voxelToMm[1].cpu(), 0, movingImage.size(1) * voxelToMm[0].cpu()]
        move = pts - pts_orig
        fig,ax=plt.subplots(1)
        ax.imshow(movingImage[0].cpu(),extent=extent,origin='lower')
        ax.set_title("vector")
        pts_orig[:, 0] = pts_orig[:, 0] * voxelToMm[0]
        pts_orig[:, 1] = pts_orig[:, 1] * voxelToMm[1]
        move[:, 0] = move[:, 0] * voxelToMm[0]
        move[:, 1] = move[:, 1] * voxelToMm[1]
        ax.quiver(pts_orig[:, 1].cpu(), pts_orig[:, 0].cpu(), move[:, 1].cpu(), move[:, 0].cpu(),color='r', angles='xy', scale_units='xy', scale=1)
        plt.show()





def thesis(device,workingDirectory,voxelToMm=None,segmentsOfInterest=None,dimension=None,slice=None):


    #BASICS: load images
    movingImageMask=torch.unsqueeze(torch.from_numpy(np.load(os.path.join(workingDirectory, "moving_mask.npy"))), dim=0).to(device=device)
    movingImage=torch.unsqueeze(torch.from_numpy(np.load(os.path.join(workingDirectory, "moving.npy"))), dim=0).to(device=device)

    indices = np.indices(movingImage.cpu()[0].size())
    pts = np.empty((np.prod(movingImage.cpu().size()), len(movingImage[0].cpu().size())))
    for i, slide in enumerate(indices):
        pts[:, i] = slide.flatten()
    pts=torch.from_numpy(pts).to(device=device).float()
    pts_orig=pts.clone()

    if dimension == 0:
        pts_orig = pts_orig[pts_orig[:, 0] == slice]
    elif dimension == 1:
        pts_orig = pts_orig[pts_orig[:, 1] == slice]
    elif dimension == 2:
        pts_orig = pts_orig[pts_orig[:, 2] == slice]
    pts=pts_orig.clone()




    inter = interpolation.nearest(device,
                                  scale=torch.ones(pts.size(1), device=device))
    ptsSegmentation = inter(pts, movingImageMask).flatten().long()

    mask = torch.zeros_like(ptsSegmentation, dtype=torch.bool)
    for segment in segmentsOfInterest:
        mask |= (ptsSegmentation == segment)
    pts=pts[mask]
    ptsSegmentation=ptsSegmentation[mask]
    pts=checkAffine(device,workingDirectory,pts,ptsSegmentation)
    pts=checkNonrigid(device,workingDirectory,pts,ptsSegmentation)
    pts_orig=pts_orig[mask]


    fig, ax = plt.subplots(ncols=2)
    movingImage = movingImage[0, :, slice, :]
    movingImageMask=movingImageMask[0, :, slice, :]
    cnts=skimage.measure.find_contours(movingImageMask.cpu().numpy())
    extent = [0, movingImage.size(1) * voxelToMm[2].cpu(), 0, movingImage.size(0) * voxelToMm[0].cpu()]

    ax[0].imshow(movingImage.cpu()*-1,extent=extent,cmap="binary",origin="lower")
    for c in cnts:
        c[:,0]=c[:,0]*voxelToMm[0].cpu().numpy()
        c[:,1]=c[:,1]*voxelToMm[2].cpu().numpy()
        ax[0].plot(c[:,1],c[:,0],c="b",linewidth=5)

    ax[1].imshow(movingImage.cpu()*-1,extent=extent,cmap="binary",origin="lower")
    for c in cnts:

        ax[1].plot(c[:,1],c[:,0],c="b")
    pts_orig0HELP=pts_orig[ptsSegmentation==0]
    pts0HELP=pts[ptsSegmentation==0]

    pts_orig1HELP=pts_orig[ptsSegmentation==1]
    pts1HELP=pts[ptsSegmentation==1]

    pts1 = pts0HELP[:, 0]
    pts2 = pts0HELP[:, 2]
    pts = torch.stack((pts1, pts2), dim=1)
    pts_orig1 = pts_orig0HELP[:, 0]
    pts_orig2 = pts_orig0HELP[:, 2]
    pts_orig = torch.stack((pts_orig1, pts_orig2), dim=1)
    move=pts-pts_orig
    pts_orig[:,0]=pts_orig[:,0]*voxelToMm[0]
    pts_orig[:,1]=pts_orig[:,1]*voxelToMm[2]
    move[:,0]=move[:,0]*voxelToMm[0]
    move[:,1]=move[:,1]*voxelToMm[2]
    ax[0].quiver(pts_orig[:, 1].cpu(), pts_orig[:, 0].cpu(), move[:, 1].cpu(), move[:, 0].cpu(),color='r', angles='xy', scale_units='xy', scale=1, width=0.0075 )

    pts1 = pts1HELP[:, 0]
    pts2 = pts1HELP[:, 2]
    pts = torch.stack((pts1, pts2), dim=1)
    pts_orig1 = pts_orig1HELP[:, 0]
    pts_orig2 = pts_orig1HELP[:, 2]
    pts_orig = torch.stack((pts_orig1, pts_orig2), dim=1)
    move=pts-pts_orig
    pts_orig[:,0]=pts_orig[:,0]*voxelToMm[0]
    pts_orig[:,1]=pts_orig[:,1]*voxelToMm[2]
    move[:,0]=move[:,0]*voxelToMm[0]
    move[:,1]=move[:,1]*voxelToMm[2]
    ax[0].quiver(pts_orig[:, 1].cpu(), pts_orig[:, 0].cpu(), move[:, 1].cpu(), move[:, 0].cpu(),color='g', angles='xy', scale_units='xy', scale=1, width=0.0075 )

    ax[1].set_xticks([])
    ax[1].set_yticks([])
    ax[0].set_xticks([])
    ax[0].set_yticks([])
    ax[0].set_ylim([70,130])
    ax[0].set_xlim([240,280])

    from matplotlib.patches import Rectangle
    square = Rectangle((240, 70), 40, 60,
                       linewidth=2, edgecolor='y', facecolor='none',linestyle='--')

    # Add the square to the plot
    ax[1].add_patch(square)



    plt.tight_layout()
    plt.show()




import torch
import re
import numpy as np
import os
import pickle
from ..import geometricTransformations
from typing import Optional, Type, Union, Tuple
from torch import Tensor

def compute_gradient_central_diff3D(F: Tensor, dx: Optional[float]=1.0, dy: Optional[float]=1.0, dz: Optional[float]=1.0)->Tuple[Tensor,Tensor,Tensor]:
    """ Computes Jacobian of dense vector fields via finite differences in 3D.
    Set the following variables
        :param F: field
        :type F: torch.Tensor
        :param dx: finite differences step in x direction
        :type dx: float
        :param dy: finite differences step in x direction
        :type dy: float
        :param dz: finite differences step in x direction
        :type dz: float
        :return dFdx: derivative in x
        :rtype dFdx: torch.Tensor
        :return dFdy: derivative in y
        :rtype dFdy: torch.Tensor
        :return dFdz: derivative in z
        :rtype dFdz: torch.Tensor
    """
    dFdx = (torch.roll(F, -1, dims=0) - torch.roll(F, 1, dims=0)) / (2 * dx)
    dFdy = (torch.roll(F, -1, dims=1) - torch.roll(F, 1, dims=1)) / (2 * dy)
    dFdz = (torch.roll(F, -1, dims=2) - torch.roll(F, 1, dims=2)) / (2 * dz)
    dFdx[0, :, :] = (F[1, :, :] - F[0, :, :]) / dx  # Forward difference
    dFdx[-1, :, :] = (F[-1, :, :] - F[-2, :, :]) / dx  # Backward difference
    dFdy[:, 0, :] = (F[:, 1, :] - F[:, 0, :]) / dy  # Forward difference
    dFdy[:, -1, :] = (F[:, -1, :] - F[:, -2, :]) / dy  # Backward difference
    dFdz[:, :, 0] = (F[:, :, 1] - F[:, :, 0]) / dz  # Forward difference
    dFdz[:, :, -1] = (F[:, :, -1] - F[:, :, -2]) / dz  # Backward difference
    return dFdx, dFdy, dFdz

def compute_gradient_central_diff2D(F: Tensor, dx: Optional[float]=1.0, dy: Optional[float]=1.0)->Tuple[Tensor,Tensor]:
    """ Computes jacobian via finite differences in 2D.
    Set the following variables
        :param F: field
        :type F: torch.Tensor
        :param dx: finite differences step in x direction
        :type dx: float
        :param dy: finite differences step in x direction
        :type dy: float
        :return dFdx: derivative in x
        :rtype dFdx: torch.tensor
        :return dFdy: derivative in y
        :rtype dFdy: torch.Tensor
    """
    dFdx = (torch.roll(F, -1, dims=0) - torch.roll(F, 1, dims=0)) / (2 * dx)
    dFdy = (torch.roll(F, -1, dims=1) - torch.roll(F, 1, dims=1)) / (2 * dy)
    dFdx[0, :] = (F[1, :] - F[0, :]) / dx  # Forward difference
    dFdx[-1, :] = (F[-1, :] - F[-2, :]) / dx  # Backward difference
    dFdy[:, 0] = (F[:, 1] - F[:, 0]) / dy  # Forward difference
    dFdy[:, -1] = (F[:, -1] - F[:, -2]) / dy  # Backward difference
    return dFdx, dFdy

def extract_segment_and_scale_JSON(file: str)->Tuple[str,str]:
    """ For automated recognition of scale and segment in nonrigid result names for json files.
        Handles single integer segments, tuple segments, and 'None' segments
        :param file: file name of result
        :type file: str
        :return scale_values: segment
        :rtype scale_values: str
        :return segment_values: scale
        :rtype segment_values: str
    """
    # Updated pattern to handle single integers, tuples, and the word 'None'
    pattern = r"transformation_nonrigid_segment_(?P<segment>\(\d+(?:, \d+)*\)|\d+|None)_scale_tensor\(\[(?P<scale>[0-9., ]+)\]\)\.json"
    # Loop over file names and extract information
    match = re.match(pattern, file)
    if match:
        segment = match.group("segment")
        scale = match.group("scale")
        # Parse the segment: handle tuple, single integer, and 'None' case
        if segment == 'None':
            segment_values = None
        elif segment.startswith("(") and segment.endswith(")"):
            # Parse tuple segment
            segment_values = tuple(int(val.strip()) for val in segment[1:-1].split(','))
        else:
            # Parse single integer segment
            segment_values = int(segment)
        # Parse the scale as a list of floats
        scale_values = [float(val.strip()) for val in scale.split(',')]

        return scale_values, segment_values
    else:
        return None, None

def extract_segment_and_scale_NPY(file: str)->Tuple[str,str]:
    """ For automated recognition of scale and segment in nonrigid result names for npy files.
        Handles single integer segments, tuple segments, and 'None' segments
        :param file: file name of result
        :type file: str
        :return scale_values: segment
        :rtype scale_values: str
        :return segment_values: scale
        :rtype segment_values: str
    """
    # Updated pattern to handle single integers, tuples, and the word 'None'
    pattern = r"transformation_nonrigid_segment_(?P<segment>\(\d+(?:, \d+)*\)|\d+|None)_scale_tensor\(\[(?P<scale>[0-9., ]+)\]\)\.npy"

    # Loop over file names and extract information
    match = re.match(pattern, file)
    if match:
        segment = match.group("segment")
        scale = match.group("scale")

        # Parse the segment: handle tuple, single integer, and 'None' case
        if segment == 'None':
            segment_values = None
        elif segment.startswith("(") and segment.endswith(")"):
            # Parse tuple segment
            segment_values = tuple(int(val.strip()) for val in segment[1:-1].split(','))
        else:
            # Parse single integer segment
            segment_values = int(segment)

        # Parse the scale as a list of floats
        scale_values = [float(val.strip()) for val in scale.split(',')]

        return scale_values, segment_values
    else:
        return None, None
def extract_scale_NPY(file: str)->str:
    """ For automated recognition of scale in nonrigid result names for npy files.
        :param file: file name of result
        :type file: str
        :return scale_values: segment
        :rtype scale_values: str
    """
    # Updated pattern to handle single integers, tuples, and the word 'None'
    pattern = r"transformation_nonrigid_scale_tensor\(\[(?P<scale>[0-9., ]+)\]\)\.npy"

    # Loop over file names and extract information
    match = re.match(pattern, file)
    if match:
        scale = match.group("scale")
        scale_values = [float(val.strip()) for val in scale.split(',')]

        return scale_values
    else:
        return None
def checkAffine(device: str,workingDirectory: str,pts: Tensor,segmentations: Optional[list]=None)->Tensor:
    """ Checks the results folder for affine transformations and applies them.
         :param device: file name of result
         :type device: str
         :param workingDirectory: workingDirectory path
         :type workingDirectory: str
         :param pts: points to be transformed
         :type pts: torch.Tensor
         :param segmentations: segmentations of interest list
         :type segmentations: list
         :return pts: transformed points
         :rtype pts: torch.Tensor
     """
    if segmentations is not None:
        for segment in torch.unique(segmentations):
            if os.path.exists(os.path.join(workingDirectory, "results", f"transformation_affine_{segment}.npy")):
                print(f" segment {segment} affine registration applied ")
                affineMat = torch.from_numpy(
                    np.load(os.path.join(workingDirectory, "results", f"transformation_affine_{segment}.npy"))).to(
                    device=device)
                ptsSegment = pts[segmentations.flatten() == segment]
                affine = geometricTransformations.affineTransformation(ptsSegment)
                ptsSegment = affine.apply(ptsSegment, affineMat)
                pts[segmentations.flatten() == segment] = ptsSegment
    else:
        if os.path.exists(os.path.join(workingDirectory, "results", "transformation_affine.npy")):
            print("overall affine registration applied")
            affineMat = torch.from_numpy(
                np.load(os.path.join(workingDirectory, "results", "transformation_affine.npy"))).to(device=device)
            affine = geometricTransformations.affineTransformation(pts)
            pts = affine.apply(pts, affineMat)
    return pts

def checkNonrigid(device: str,workingDirectory: str,pts: Tensor,segmentations: Optional[list]=None)-> Tensor:
    """ Checks the results folder for nonrigied transformations and applies them.
         :param device: file name of result
         :type device: str
         :param workingDirectory: workingDirectory path
         :type workingDirectory: str
         :param pts: points to be transformed
         :type pts: torch.Tensor
         :param segmentations: segmentations of interest list
         :type segmentations: list
         :return pts: transformed points
         :rtype pts: torch.Tensor
     """
    if segmentations is not None:
        files = os.listdir(os.path.join(workingDirectory, "results"))
        filtered_files = [file for file in files if
                          file.startswith("transformation_nonrigid_segment") and file.endswith(".npy")]
        sorted_files = sorted(filtered_files,
                              key=lambda s: (extract_segment_and_scale_NPY(s)[0], extract_segment_and_scale_NPY(s)[1]),
                              reverse=True)
        for f in sorted_files:
            velocityField = torch.from_numpy(np.load(os.path.join(workingDirectory, "results", f))).to(device=device)
            scale,segment=extract_segment_and_scale_NPY(f)


            ptsSegment = pts[segmentations.flatten() == segment]

            if ptsSegment.size(0)>0:
                pass
                print(f" segment {segment} deformable registration applied at scale {scale}")

            scale = torch.tensor(scale).to(device=device)
            with open(os.path.join(workingDirectory, "results",
                                   f"class_transformation_nonrigid_segment_{segment}_scale_" + str(
                                       scale.cpu()) + ".pkl"), 'rb') as input_file:
                nrDeformation = pickle.load(input_file)
            ptsSegment = nrDeformation.apply(ptsSegment, velocityField)
            pts[segmentations.flatten() == segment] = ptsSegment

    else:
        files = os.listdir(os.path.join(workingDirectory, "results"))
        filtered_files = [file for file in files if
                          file.startswith("transformation_nonrigid") and file.endswith(".npy")]
        sorted_files = sorted(filtered_files,
                              key=lambda s: (extract_scale_NPY(s)[0]),
                              reverse=True)
        for f in sorted_files:
            velocityField = torch.from_numpy(np.load(os.path.join(workingDirectory, "results", f))).to(device=device)
            scale=extract_scale_NPY(f)
            print(f" general deformable registration applied at scale {scale}")

            scale = torch.tensor(scale).to(device=device)
            with open(os.path.join(workingDirectory, "results",
                                   "class_transformation_nonrigid_scale_" + str(scale.cpu()) + ".pkl"),
                      'rb') as input_file:
                nrDeformation = pickle.load(input_file)
            pts = nrDeformation.apply(pts, velocityField)
    return pts

def checkAffineInverse(device: str,workingDirectory: str,pts: Tensor,segmentations: Optional[list]=None)->Tensor:
    """ Checks the results folder for affine transformations and applies the inverse.
         :param device: file name of result
         :type device: str
         :param workingDirectory: workingDirectory path
         :type workingDirectory: str
         :param pts: points to be transformed
         :type pts: torch.Tensor
         :param segmentations: segmentations of interest list
         :type segmentations: list
         :return pts: transformed points
         :rtype pts: torch.Tensor
     """
    if segmentations is not None:
        for segment in torch.unique(segmentations):
            if os.path.exists(os.path.join(workingDirectory, "results", f"transformation_affine_{segment}.npy")):
                #print(f" segment {segment} inverse affine registration applied ")
                affineMat = torch.from_numpy(
                    np.load(os.path.join(workingDirectory, "results", f"transformation_affine_{segment}.npy"))).to(
                    device=device)
                affineMat=torch.linalg.inv(affineMat)
                ptsSegment = pts[segmentations.flatten() == segment]
                affine = geometricTransformations.affineTransformation(ptsSegment)
                ptsSegment = affine.apply(ptsSegment, affineMat)
                pts[segmentations.flatten() == segment] = ptsSegment
    else:
        if os.path.exists(os.path.join(workingDirectory, "results", "transformation_affine.npy")):
            print("overall inverse affine registration applied")
            affineMat = torch.from_numpy(
                np.load(os.path.join(workingDirectory, "results", "transformation_affine.npy"))).to(device=device)
            affineMat = torch.linalg.inv(affineMat)

            affine = geometricTransformations.affineTransformation(pts)
            pts = affine.apply(pts, affineMat)
    return pts

def checkNonrigidInverse(device: str,workingDirectory: str,pts: Tensor,segmentations: Optional[list]=None)-> Tensor:
    """ Checks the results folder for nonrigied transformations and applies the inverse.
         :param device: file name of result
         :type device: str
         :param workingDirectory: workingDirectory path
         :type workingDirectory: str
         :param pts: points to be transformed
         :type pts: torch.Tensor
         :param segmentations: segmentations of interest list
         :type segmentations: list
         :return pts: transformed points
         :rtype pts: torch.Tensor
     """
    if segmentations is not None:
        files = os.listdir(os.path.join(workingDirectory, "results"))
        filtered_files = [file for file in files if
                          file.startswith("transformation_nonrigid_segment") and file.endswith(".npy")]
        sorted_files = sorted(filtered_files,
                              key=lambda s: (extract_segment_and_scale_NPY(s)[0], extract_segment_and_scale_NPY(s)[1]),
                              reverse=False)
        for f in sorted_files:
            velocityField = torch.from_numpy(np.load(os.path.join(workingDirectory, "results", f))).to(device=device)
            velocityField = torch.flip(velocityField, [0]) * -1
            scale,segment=extract_segment_and_scale_NPY(f)


            ptsSegment = pts[segmentations.flatten() == segment]
            if ptsSegment.size(0)>0:
                pass
                #print(f" segment {segment} deformable registration applied at scale {scale}")
            scale = torch.tensor(scale).to(device=device)
            with open(os.path.join(workingDirectory, "results",
                                   f"class_transformation_nonrigid_segment_{segment}_scale_" + str(
                                       scale.cpu()) + ".pkl"), 'rb') as input_file:
                nrDeformation = pickle.load(input_file)
            ptsSegment = nrDeformation.apply(ptsSegment, velocityField)
            pts[segmentations.flatten() == segment] = ptsSegment

    else:
        files = os.listdir(os.path.join(workingDirectory, "results"))
        filtered_files = [file for file in files if
                          file.startswith("transformation_nonrigid") and file.endswith(".npy")]
        sorted_files = sorted(filtered_files,
                              key=lambda s: (extract_scale_NPY(s)[0]),
                              reverse=False)
        for f in sorted_files:
            velocityField = torch.from_numpy(np.load(os.path.join(workingDirectory, "results", f))).to(device=device)
            velocityField = torch.flip(velocityField, [0]) * -1
            scale=extract_scale_NPY(f)
            print(f" general deformable registration applied at scale {scale}")

            scale = torch.tensor(scale).to(device=device)
            with open(os.path.join(workingDirectory, "results",
                                   "class_transformation_nonrigid_scale_" + str(scale.cpu()) + ".pkl"),
                      'rb') as input_file:
                nrDeformation = pickle.load(input_file)
            pts = nrDeformation.apply(pts, velocityField)
    return pts
